# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import filecmp
import pickle

from ploigos_step_runner import StepResult, WorkflowResult
from ploigos_step_runner.exceptions import StepRunnerException
from testfixtures import TempDirectory
from tests.helpers.base_test_case import BaseTestCase


def setup_test():
    step_result1 = StepResult('step1', 'sub1', 'implementer1')
    step_result1.add_artifact('artifact1', 'value1', 'description1')
    step_result1.add_artifact('artifact2', 'value2', 'description2')
    step_result1.add_artifact('artifact3', 'value3')
    step_result1.add_artifact('artifact4', False)
    step_result1.add_artifact('same-artifact-all-env-and-no-env', 'result1')

    step_result1.add_evidence('evidence1', 'value1', 'description1')
    step_result1.add_evidence('evidence2', 'value2', 'description2')
    step_result1.add_evidence('evidence3', 'value3')
    step_result1.add_evidence('evidence4', False)
    step_result1.add_evidence('same-evidence-all-env-and-no-env', 'result1')


    step_result2 = StepResult('step2', 'sub2', 'implementer2')
    step_result2.add_artifact('artifact1', True)
    step_result2.add_artifact('artifact2', False)
    step_result2.add_artifact('artifact5', 'value5')

    step_result2.add_evidence('evidence1', True)
    step_result2.add_evidence('evidence2', False)
    step_result2.add_evidence('evidence5', 'value5')

    step_result3 = StepResult('deploy', 'deploy-sub', 'helm', 'dev')
    step_result3.add_artifact('same-artifact-diff-env', 'value-dev-env')
    step_result3.add_artifact('unique-artifact-to-step-and-environment-1', 'value1-dev-env')
    step_result3.add_artifact('same-artifact-all-env-and-no-env', 'result3-dev-env')

    step_result3.add_evidence('same-evidence-diff-env', 'value-dev-env')
    step_result3.add_evidence('unique-evidence-to-step-and-environment-1', 'value1-dev-env')
    step_result3.add_evidence('same-evidence-all-env-and-no-env', 'result3-dev-env')

    step_result4 = StepResult('deploy', 'deploy-sub', 'helm', 'test')
    step_result4.add_artifact('artifact1', True)
    step_result4.add_artifact('artifact2', False)
    step_result4.add_artifact('artifact5', 'value5')
    step_result4.add_artifact('same-artifact-diff-env', 'value-test-env')
    step_result4.add_artifact('unique-artifact-to-step-and-environment-2', 'value2-test-env')
    step_result4.add_artifact('same-artifact-all-env-and-no-env', 'result4-test-env')

    step_result4.add_evidence('evidence1', True)
    step_result4.add_evidence('evidence2', False)
    step_result4.add_evidence('evidence5', 'value5')
    step_result4.add_evidence('same-evidence-diff-env', 'value-test-env')
    step_result4.add_evidence('unique-evidence-to-step-and-environment-2', 'value2-test-env')
    step_result4.add_evidence('same-evidence-all-env-and-no-env', 'result4-test-env')

    wfr = WorkflowResult()
    wfr.add_step_result(step_result1)
    wfr.add_step_result(step_result2)
    wfr.add_step_result(step_result3)
    wfr.add_step_result(step_result4)

    return wfr


def setup_test_sub_steps():
    step_result1 = StepResult('step1', 'sub1', 'implementer1')
    step_result1.add_artifact('artifact1', 'value1', 'description1')
    step_result1.add_artifact('artifact2', 'value2', 'description2')
    step_result1.add_artifact('artifact3', 'value3')
    step_result1.add_artifact('artifact4', False)

    step_result1.add_evidence('evidence1', 'value1', 'description1')
    step_result1.add_evidence('evidence2', 'value2', 'description2')
    step_result1.add_evidence('evidence3', 'value3')
    step_result1.add_evidence('evidence4', False)

    step_result2 = StepResult('step1', 'sub2', 'implementer2')
    step_result2.add_artifact('artifact1', True)
    step_result2.add_artifact('artifact2', False)
    step_result2.add_artifact('artifact5', 'value5')

    step_result2.add_evidence('evidence1', True)
    step_result2.add_evidence('evidence2', False)
    step_result2.add_evidence('evidence5', 'value5')

    wfr = WorkflowResult()
    wfr.add_step_result(step_result1)
    wfr.add_step_result(step_result2)

    return wfr


class TestStepWorkflowResultTest(BaseTestCase):
    """
    :return:
    """

    def test_get_artifact_value_without_step(self):
        wfr = setup_test()

        expected_artifact = True
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact1')
        )

        expected_artifact = 'value5'
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact5')
        )

        expected_artifact = False
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact4')
        )

        expected_artifact = None
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='bad')
        )

    def test_get_evidence_value_without_step(self):
        wfr = setup_test()

        expected_evidence = 'value1'
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence1')
        )

        expected_evidence = 'value5'
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence5')
        )

        expected_evidence = False
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence4')
        )

        expected_evidence = None
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='bad')
        )

    def test_get_artifact_value_with_step(self):
        wfr = setup_test()

        expected_artifact = 'value1'
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact1', step_name='step1')
        )
        expected_artifact = False
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact2', step_name='step2')
        )
        expected_artifact = None
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact2', step_name='bad')
        )

    def test_get_evidence_value_with_step(self):
        wfr = setup_test()

        expected_evidence = 'value1'
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence1', step_name='step1')
        )
        expected_evidence = False
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence2', step_name='step2')
        )
        expected_evidence = None
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence2', step_name='bad')
        )

    def test_get_artifact_value_with_step_sub_step(self):
        wfr = setup_test()

        expected_artifact = 'value1'
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact1', step_name='step1', sub_step_name='sub1')
        )
        expected_artifact = False
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact2', step_name='step2', sub_step_name='sub2')
        )
        expected_artifact = None
        self.assertEqual(
            expected_artifact,
            wfr.get_artifact_value(artifact='artifact2', step_name='step2', sub_step_name='bad')
        )

    def test_get_evidence_value_with_step_sub_step(self):
        wfr = setup_test()

        expected_evidence = 'value1'
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence1', step_name='step1', sub_step_name='sub1')
        )
        expected_evidence = False
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence2', step_name='step2', sub_step_name='sub2')
        )
        expected_evidence = None
        self.assertEqual(
            expected_evidence,
            wfr.get_evidence_value(evidence='evidence2', step_name='step2', sub_step_name='bad')
        )

    def test_get_artifact_value_in_step_executed_in_different_environments_and_no_environment(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-all-env-and-no-env'
            ),
            'result4-test-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-all-env-and-no-env',
                environment='dev'
            ),
            'result3-dev-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-all-env-and-no-env',
                environment='test'
            ),
            'result4-test-env'
        )

    def test_get_evidence_value_in_step_executed_in_different_environments_and_no_environment(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-all-env-and-no-env'
            ),
            'result1'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-all-env-and-no-env',
                environment='dev'
            ),
            'result3-dev-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-all-env-and-no-env',
                environment='test'
            ),
            'result4-test-env'
        )

    def test_get_artifact_value_in_step_executed_in_different_environments(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-diff-env'
            ),
            'value-test-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-diff-env',
                environment='dev'
            ),
            'value-dev-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='same-artifact-diff-env',
                environment='test'
            ),
            'value-test-env'
        )

    def test_get_evidence_value_in_step_executed_in_different_environments(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-diff-env'
            ),
            'value-dev-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-diff-env',
                environment='dev'
            ),
            'value-dev-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='same-evidence-diff-env',
                environment='test'
            ),
            'value-test-env'
        )

    def test_get_artifact_value_unique_to_environment(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-1'
            ),
            'value1-dev-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-1',
                environment='dev'
            ),
            'value1-dev-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-1',
                environment='test'
            ),
            None
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-2'
            ),
            'value2-test-env'
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-2',
                environment='dev'
            ),
            None
        )

        self.assertEqual(
            wfr.get_artifact_value(
                artifact='unique-artifact-to-step-and-environment-2',
                environment='test'
            ),
            'value2-test-env'
        )

    def test_get_evidence_value_unique_to_environment(self):
        wfr = setup_test()

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-1'
            ),
            'value1-dev-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-1',
                environment='dev'
            ),
            'value1-dev-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-1',
                environment='test'
            ),
            None
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-2'
            ),
            'value2-test-env'
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-2',
                environment='dev'
            ),
            None
        )

        self.assertEqual(
            wfr.get_evidence_value(
                evidence='unique-evidence-to-step-and-environment-2',
                environment='test'
            ),
            'value2-test-env'
        )

    def test_add_step_result_duplicate_no_env(self):
        duplicate_step_result = StepResult('step1', 'sub1', 'implementer1')
        duplicate_step_result.add_artifact('newartifact', 'newvalue', 'newdescription')
        duplicate_step_result.add_evidence('newevidence', 'newvalue', 'newdescription')

        wfr = setup_test()

        with self.assertRaisesRegex(
            StepRunnerException,
            r"Can not add duplicate StepResult for step \(step1\),"
            r" sub step \(sub1\), and environment \(None\)."
        ):
            wfr.add_step_result(duplicate_step_result)

    def test_add_step_result_duplicate_with_env(self):
        duplicate_step_result = StepResult('deploy', 'deploy-sub', 'helm', 'dev')
        duplicate_step_result.add_artifact('newartifact', 'newvalue', 'newdescription')
        duplicate_step_result.add_evidence('newevidence', 'newvalue', 'newdescription')

        wfr = setup_test()

        with self.assertRaisesRegex(
            StepRunnerException,
            r"Can not add duplicate StepResult for step \(deploy\),"
            r" sub step \(deploy-sub\), and environment \(dev\)."
        ):
            wfr.add_step_result(duplicate_step_result)

    def test_add_step_result_exception(self):
        wfr = setup_test()

        with self.assertRaisesRegex(
                StepRunnerException,
                r"expect StepResult instance type"):
            wfr.add_step_result(step_result="bad")

    def test_write_results_to_yml_file(self):
        expected_yml_result = """step-runner-results:
    step1:
        sub1:
            artifacts:
            -   description: description1
                name: artifact1
                value: value1
            -   description: description2
                name: artifact2
                value: value2
            -   description: ''
                name: artifact3
                value: value3
            -   description: ''
                name: artifact4
                value: false
            evidence:
            -   description: description1
                name: evidence1
                value: value1
            -   description: description2
                name: evidence2
                value: value2
            -   description: ''
                name: evidence3
                value: value3
            -   description: ''
                name: evidence4
                value: false
            message: ''
            sub-step-implementer-name: implementer1
            success: true
        sub2:
            artifacts:
            -   description: ''
                name: artifact1
                value: true
            -   description: ''
                name: artifact2
                value: false
            -   description: ''
                name: artifact5
                value: value5
            evidence:
            -   description: ''
                name: evidence1
                value: true
            -   description: ''
                name: evidence2
                value: false
            -   description: ''
                name: evidence5
                value: value5
            message: ''
            sub-step-implementer-name: implementer2
            success: true
"""
        wfr = setup_test_sub_steps()
        with TempDirectory() as temp_dir:
            yml_file = temp_dir.path + '/test-results.yml'
            wfr.write_results_to_yml_file(yml_file)

            expected_yml_file = temp_dir.path + '/test-expected-results.yml'

            with open(yml_file, 'r') as actual:
                self.assertEqual(actual.read(), expected_yml_result)

    def test_write_results_to_yml_file_exception(self):
        wfr = setup_test()

        with self.assertRaises(RuntimeError):
            wfr.write_results_to_yml_file('/NotAStepResult/dir/test.yml')

    def test_write_results_to_json_file(self):
        expected_json_result = """{
    "step-runner-results": {
        "step1": {
            "sub1": {
                "sub-step-implementer-name": "implementer1",
                "success": true,
                "message": "",
                "artifacts": [
                    {
                        "name": "artifact1",
                        "value": "value1",
                        "description": "description1"
                    },
                    {
                        "name": "artifact2",
                        "value": "value2",
                        "description": "description2"
                    },
                    {
                        "name": "artifact3",
                        "value": "value3",
                        "description": ""
                    },
                    {
                        "name": "artifact4",
                        "value": false,
                        "description": ""
                    },
                    {
                        "name": "same-artifact-all-env-and-no-env",
                        "value": "result1",
                        "description": ""
                    }
                ],
                "evidence": [
                    {
                        "name": "evidence1",
                        "value": "value1",
                        "description": "description1"
                    },
                    {
                        "name": "evidence2",
                        "value": "value2",
                        "description": "description2"
                    },
                    {
                        "name": "evidence3",
                        "value": "value3",
                        "description": ""
                    },
                    {
                        "name": "evidence4",
                        "value": false,
                        "description": ""
                    },
                    {
                        "name": "same-evidence-all-env-and-no-env",
                        "value": "result1",
                        "description": ""
                    }
                ]
            }
        },
        "step2": {
            "sub2": {
                "sub-step-implementer-name": "implementer2",
                "success": true,
                "message": "",
                "artifacts": [
                    {
                        "name": "artifact1",
                        "value": true,
                        "description": ""
                    },
                    {
                        "name": "artifact2",
                        "value": false,
                        "description": ""
                    },
                    {
                        "name": "artifact5",
                        "value": "value5",
                        "description": ""
                    }
                ],
                "evidence": [
                    {
                        "name": "evidence1",
                        "value": true,
                        "description": ""
                    },
                    {
                        "name": "evidence2",
                        "value": false,
                        "description": ""
                    },
                    {
                        "name": "evidence5",
                        "value": "value5",
                        "description": ""
                    }
                ]
            }
        },
        "dev": {
            "deploy": {
                "deploy-sub": {
                    "sub-step-implementer-name": "helm",
                    "success": true,
                    "message": "",
                    "artifacts": [
                        {
                            "name": "same-artifact-diff-env",
                            "value": "value-dev-env",
                            "description": ""
                        },
                        {
                            "name": "unique-artifact-to-step-and-environment-1",
                            "value": "value1-dev-env",
                            "description": ""
                        },
                        {
                            "name": "same-artifact-all-env-and-no-env",
                            "value": "result3-dev-env",
                            "description": ""
                        }
                    ],
                    "evidence": [
                        {
                            "name": "same-evidence-diff-env",
                            "value": "value-dev-env",
                            "description": ""
                        },
                        {
                            "name": "unique-evidence-to-step-and-environment-1",
                            "value": "value1-dev-env",
                            "description": ""
                        },
                        {
                            "name": "same-evidence-all-env-and-no-env",
                            "value": "result3-dev-env",
                            "description": ""
                        }
                    ]
                }
            }
        },
        "test": {
            "deploy": {
                "deploy-sub": {
                    "sub-step-implementer-name": "helm",
                    "success": true,
                    "message": "",
                    "artifacts": [
                        {
                            "name": "artifact1",
                            "value": true,
                            "description": ""
                        },
                        {
                            "name": "artifact2",
                            "value": false,
                            "description": ""
                        },
                        {
                            "name": "artifact5",
                            "value": "value5",
                            "description": ""
                        },
                        {
                            "name": "same-artifact-diff-env",
                            "value": "value-test-env",
                            "description": ""
                        },
                        {
                            "name": "unique-artifact-to-step-and-environment-2",
                            "value": "value2-test-env",
                            "description": ""
                        },
                        {
                            "name": "same-artifact-all-env-and-no-env",
                            "value": "result4-test-env",
                            "description": ""
                        }
                    ],
                    "evidence": [
                        {
                            "name": "evidence1",
                            "value": true,
                            "description": ""
                        },
                        {
                            "name": "evidence2",
                            "value": false,
                            "description": ""
                        },
                        {
                            "name": "evidence5",
                            "value": "value5",
                            "description": ""
                        },
                        {
                            "name": "same-evidence-diff-env",
                            "value": "value-test-env",
                            "description": ""
                        },
                        {
                            "name": "unique-evidence-to-step-and-environment-2",
                            "value": "value2-test-env",
                            "description": ""
                        },
                        {
                            "name": "same-evidence-all-env-and-no-env",
                            "value": "result4-test-env",
                            "description": ""
                        }
                    ]
                }
            }
        }
    }
}"""
        wfr = setup_test()
        with TempDirectory() as temp_dir:
            json_file_path = temp_dir.path + '/test-results.json'
            wfr.write_results_to_json_file(json_file_path)

            with open(json_file_path, 'r') as actual_json_file:
                json_file_contents = actual_json_file.read()
                self.assertEqual(json_file_contents, expected_json_result)

    def test_write_results_to_json_file_exception(self):
        wfr = setup_test()

        # the sub-folders will be created
        # with TempDirectory() as temp_dir:
        #     wfr.write_results_to_json_file(f'{temp_dir.path}/NotAStepResult/dir/test.json')
        with self.assertRaises(
                RuntimeError):
            wfr.write_results_to_json_file('/NotAStepResult/dir/test.json')

    def test_load_from_pickle_file_no_file(self):
        pickle_wfr = WorkflowResult.load_from_pickle_file('test.pkl')
        expected_wfr = WorkflowResult()
        self.assertEqual(
            pickle_wfr._WorkflowResult__get_all_step_results_dict(),
            expected_wfr._WorkflowResult__get_all_step_results_dict()
        )

    def test_load_from_pickle_file_empty_file(self):
        with TempDirectory() as temp_dir:
            pickle_file = temp_dir.path + '/test.pkl'
            open(pickle_file, 'a').close()
            pickle_wfr = WorkflowResult.load_from_pickle_file(pickle_file)
            expected_wfr = WorkflowResult()
            self.assertEqual(
                pickle_wfr._WorkflowResult__get_all_step_results_dict(),
                expected_wfr._WorkflowResult__get_all_step_results_dict()
            )

    def test_load_from_pickle_file_no_workflowresult(self):
        with TempDirectory() as temp_dir:
            pickle_file = temp_dir.path + '/test.pkl'

            not_wfr = {"step1": "value1", "step2": "value2"}
            with open(pickle_file, 'wb') as file:
                pickle.dump(not_wfr, file)

            with self.assertRaisesRegex(
                    StepRunnerException,
                    f'error {pickle_file} has invalid data'):
                WorkflowResult.load_from_pickle_file(pickle_file)

    def test_load_from_pickle_file_yes_workflow_result(self):
        with TempDirectory() as temp_dir:
            pickle_file = temp_dir.path + '/test.pkl'
            expected_wfr = setup_test()
            expected_wfr.write_to_pickle_file(pickle_file)
            pickle_wfr = WorkflowResult.load_from_pickle_file(pickle_file)

    def test_load_from_pickle_file_exception(self):
        with TempDirectory() as temp_dir:
            pickle_file_name = temp_dir.path + '/test.pkl'

            pickle_file = open(pickle_file_name, 'w+')
            pickle_file.write("This is not a Workflow Result.")
            pickle_file.close()

            with self.assertRaises(
                    StepRunnerException):
                WorkflowResult.load_from_pickle_file(pickle_file)

    def test_write_to_pickle_file(self):
        wfr = setup_test()
        with self.assertRaises(
                RuntimeError):
            wfr.write_to_pickle_file(None)

    def test_write_to_pickle_file_with_merge(self):
        with TempDirectory() as temp_dir:
            pickle_file = temp_dir.path + '/test.pkl'
            on_disk_wfr = setup_test()
            on_disk_wfr.write_to_pickle_file(pickle_file)

            in_mem_wfr1 = WorkflowResult.load_from_pickle_file(pickle_file)
            in_mem_wfr2 = WorkflowResult.load_from_pickle_file(pickle_file)

            sr1 = StepResult('compliance-scan', 'scan with stackrox', 'stackrox', 'prod')
            sr1.add_artifact('compliance-scan-result', 'pass')

            sr2 = StepResult('vulnerability-scan', 'scan with stackrox', 'stackrox', 'prod')
            sr2.add_artifact('vulnerability-scan-result', 'fail')

            in_mem_wfr1.add_step_result(sr1)
            in_mem_wfr2.add_step_result(sr2)

            in_mem_wfr1.merge_with_pickle_file(pickle_file)
            in_mem_wfr1.write_to_pickle_file(pickle_file)

            in_mem_wfr2.merge_with_pickle_file(pickle_file)
            in_mem_wfr2.write_to_pickle_file(pickle_file)

            resulting_wfr = WorkflowResult.load_from_pickle_file(pickle_file)

            # ensure:
            #   - both our new StepResults are present
            #   - all StepResults from the original pickle file are present
            self.assertIn(sr1, resulting_wfr.workflow_list)
            self.assertIn(sr2, resulting_wfr.workflow_list)
            for sr in on_disk_wfr.workflow_list:
                self.assertIn(sr, resulting_wfr.workflow_list)

    def test_merge_matching_stepresults_on_pickle(self):
        """When we pickle to disk, make sure we merge the artifacts
        and evidence from in-memory StepResult instances with those on disk.

        Whatever is in memory wins.
        """
        with TempDirectory() as temp_dir:
            pickle_file = temp_dir.path + '/test.pkl'

            expected_wfr = setup_test()
            expected_wfr.write_to_pickle_file(pickle_file)

            # now adjust our expected WorkflowResult to include the
            # amended evidence/artifacts we expect to see. All other
            # StepResults and their artifacts/evidence are left the same.
            expected_sr_no_env = expected_wfr.get_step_result('step1', 'sub1')
            expected_sr_no_env.add_artifact('artifact1', 'changed-value1', 'description1')
            expected_sr_no_env.add_evidence('evidence1', 'changed-value1', 'description1')

            expected_sr_env = expected_wfr.get_step_result('deploy', 'deploy-sub', 'dev')
            expected_sr_env.add_artifact('artifact1', 'changed-value1')
            expected_sr_env.add_evidence('evidence1', 'changed-value1', 'description1')

            # now build our 'in memory' WFR that we're going to write
            # to disk. this has the amended evidence/artifacts.
            # one StepResult with environment, one without.
            in_mem_wfr = WorkflowResult()

            sr_no_env = StepResult('step1', 'sub1', 'implementer1')
            sr_no_env.add_artifact('artifact1', 'changed-value1', 'description1')
            sr_no_env.add_evidence('evidence1', 'changed-value1', 'description1')

            sr_env = StepResult('deploy', 'deploy-sub', 'helm', 'dev')
            sr_env.add_artifact('artifact1', 'changed-value1')
            sr_env.add_evidence('evidence1', 'changed-value1', 'description1')

            in_mem_wfr.add_step_result(sr_no_env)
            in_mem_wfr.add_step_result(sr_env)

            # write in-memory to disk post-merging
            in_mem_wfr.merge_with_pickle_file(pickle_file)
            in_mem_wfr.write_to_pickle_file(pickle_file)

            resulting_wfr = WorkflowResult.load_from_pickle_file(pickle_file)
            resulting_workflow_list = resulting_wfr.workflow_list

            for expected_result in expected_wfr.workflow_list:
                merged_result = resulting_wfr.get_step_result(
                                    step_name=expected_result.step_name,
                                    sub_step_name=expected_result.sub_step_name,
                                    environment=expected_result.environment
                                  )
                self.assertEqual(expected_result, merged_result)
                resulting_workflow_list.remove(expected_result)

            # nothing duplicate or extra should be in the result
            self.assertEqual(len(resulting_workflow_list), 0)



"""`StepImplementer` for the `push-container-image` step using Skopeo.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key  | Required? | Default | Description
-------------------|-----------|---------|-----------
`container-image-version` \
                   | Yes       |         | Version to use when pushing the container image
`organization`     | Yes       |         | Used in creating the container image push tag
`application_name` | Yes       |         | Used in creating the container image push tag
`service_name`     | Yes       |         | Used in creating the container image push tag
`destination-url`  | Yes       |         | Container image repository destination to push image \
                                           to. <br/> \
                                           Should not include the repository type.
`[source-tls,verify, src-tls-verify]` \
                   | Yes       | `True`  | Whether to verify TLS when pulling source image.
`dest-tls-verify`  | Yes       | `True`  | Whether to verify TLS when pushing destination image.
`[container-image-pull-tag, container-image-tag]` \
                   | Yes       |         | Container image tag of image to push to \
                                           `destination-url`.
`[container-image-pull-repository-type, container-image-repository-type]` \
                   | Yes       | 'containers-storage:' \
                                         | Container repository type for the pull image source. \
                                           See https://github.com/containers/skopeo for valid \
                                           options.
`[container-image-push-repository-type, container-image-repository-type]` \
                   | Yes       | 'docker://' \
                                         | Container repository type for the push image source. \
                                           See https://github.com/containers/skopeo for valid \
                                           options.
`containers-config-auth-file` \
                   | No        |         | Path to the container registry authentication file \
                                           to use for container registry authentication. \
                                           If one is not provided one will be created in the \
                                           working directory.
`container-registries` \
                   | False     |         | Hash of container registries to authenticate with.


Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key            | Description
-------------------------------|------------
`container-image-registry-uri` | Registry URI poriton of the container image tag \
                                 of the pushed container image.
`container-image-registry-organization` \
                               | Organization portion of the container image tag \
                                 of the pushed container image.
`container-image-repository`   | Repository portion of the container image tag \
                                 of the pushed container image.
`container-image-name`         | Another way to reference the \
                                 repository portion of the container image tag \
                                 of the pushed container image.
`container-image-version`      | Version portion of the container image tag \
                                 of the pushed container image.
`container-image-tag`          | Full container image tag of the pushed container, \
                                 including the registry URI. <br/> \
                                 Takes the form of: \
                                    `container-image-registry-organization/container-image-repository:container-image-version`
`container-image-short-tag`    | Short container image tag of the pushed container image, \
                                 excluding the registry URI. <br/> \
                                 Takes the form of: \
                                    `container-image-registry-uri/container-image-registry-organization/container-image-repository:container-image-version`

""" # pylint: disable=line-too-long

import os
import sys
from distutils import util

import sh
from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.containers import container_registries_login

DEFAULT_CONFIG = {
    'src-tls-verify': True,
    'dest-tls-verify': True,
    'container-image-pull-repository-type': 'containers-storage:',
    'container-image-push-repository-type': 'docker://'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'destination-url',
    ['source-tls,verify', 'src-tls-verify'],
    'dest-tls-verify',
    'service-name',
    'application-name',
    'organization',

    # being flexible for different use cases of proceeding steps
    ['container-image-pull-tag', 'container-image-tag'],
    ['container-image-pull-repository-type', 'container-image-repository-type'],
    ['container-image-push-repository-type', 'container-image-repository-type']
]

class Skopeo(StepImplementer):
    """`StepImplementer` for the `push-container-image` step using Skopeo.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # get config
        image_pull_tag = self.get_value(['container-image-pull-tag', 'container-image-tag'])
        pull_repository_type = self.get_value([
            'container-image-pull-repository-type',
            'container-image-repository-type'
        ])
        push_repository_type = self.get_value([
            'container-image-push-repository-type',
            'container-image-repository-type'
        ])
        dest_tls_verify = self.get_value('dest-tls-verify')
        if isinstance(dest_tls_verify, str):
            dest_tls_verify = bool(util.strtobool(dest_tls_verify))
        source_tls_verify = self.get_value(['source-tls-verify', 'src-tls-verify'])
        if isinstance(source_tls_verify, str):
            source_tls_verify = bool(util.strtobool(source_tls_verify))

        # create push tag
        image_version = self.get_value('container-image-version')
        if image_version is None:
            image_version = 'latest'
            print('No image tag version found in metadata. Using latest')
        image_version = image_version.lower()
        image_registry_uri = self.get_value('destination-url')
        image_registry_organization = self.get_value('organization')
        image_repository = f"{self.get_value('application-name')}-{self.get_value('service-name')}"
        image_push_short_tag = f"{image_registry_organization}/{image_repository}:{image_version}"
        image_push_tag = f"{image_registry_uri}/{image_push_short_tag}"

        try:
            # login to any provider container registries
            # NOTE: important to specify the auth file because depending on the context this is
            #       being run in python process may not have permissions to default location
            containers_config_auth_file = self.get_value('containers-config-auth-file')
            if not containers_config_auth_file:
                containers_config_auth_file = os.path.join(
                    self.work_dir_path,
                    'container-auth.json'
                )
            container_registries_login(
                registries=self.get_value('container-registries'),
                containers_config_auth_file=containers_config_auth_file,
                containers_config_tls_verify=dest_tls_verify
            )

            # push image
            sh.skopeo.copy( # pylint: disable=no-member
                f"--src-tls-verify={str(source_tls_verify).lower()}",
                f"--dest-tls-verify={str(dest_tls_verify).lower()}",
                f"--authfile={containers_config_auth_file}",
                f'{pull_repository_type}{image_pull_tag}',
                f'{push_repository_type}{image_push_tag}',
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )
        except sh.ErrorReturnCode as error:
            step_result.success = False
            step_result.message = f'Error pushing container image ({image_pull_tag}) ' \
                f' to tag ({image_push_tag}) using skopeo: {error}'

        # add artifacts
        step_result.add_artifact(
            name='container-image-registry-uri',
            value=image_registry_uri,
            description='Registry URI poriton of the container image tag' \
                ' of the pushed container image.'
        )
        step_result.add_artifact(
            name='container-image-registry-organization',
            value=image_registry_organization,
            description='Organization portion of the container image tag' \
                ' of the pushed container image.'
        )
        step_result.add_artifact(
            name='container-image-repository',
            value=image_repository,
            description='Repository portion of the container image tag' \
                ' of the pushed container image.'
        )
        step_result.add_artifact(
            name='container-image-name',
            value=image_repository,
            description='Another way to reference the' \
                ' repository portion of the container image tag of the pushed container image.'
        )
        step_result.add_artifact(
            name='container-image-version',
            value=image_version,
            description='Version portion of the container image tag of the pushed container image.'
        )
        step_result.add_artifact(
            name='container-image-tag',
            value=image_push_tag,
            description='Full container image tag of the pushed container,' \
                ' including the registry URI.'
        )
        step_result.add_artifact(
            name='container-image-short-tag',
            value=image_push_short_tag,
            description='Short container image tag of the pushed container image,' \
                ' excluding the registry URI.'
        )

        return step_result



"""`StepImplementer` for the `tag-source` step using Git.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key | Required?          | Default              | Description
------------------|--------------------|----------------------|-----------
`url`             | Yes                | git url returned by \
                                         `git config --get \
                                         remote.origin.url` \
                                         in container         | This is the url for the \
                                                                git server.
`git-username`    | Yes (if url \
                    is http or https)  |                      | This is the username to use. \
                                                                Due to security concerns this \
                                                                should only be only provided by \
                                                                the runtime configuration
`git-password`    | Yes (if url \
                    is http or https)  |                      | This is the password to use. \
                                                                Due to security concerns this \
                                                                should only be only provided by \
                                                                the runtime configuration
`version`         | No                 | `latest`             | Semantic version, \
                                                                if not found this step \
                                                                will use `latest` as the version \
                                                                to perform the git tag with

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`tag`               | This is the value that was used to tag the source.
"""
import re
import sys
from io import StringIO

import sh
from ploigos_step_runner import StepImplementer
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner import StepResult

DEFAULT_CONFIG = {}

AUTHENTICATION_CONFIG = {
    'git-username': None,
    'git-password': None
}


class Git(StepImplementer):
    """StepImplementer for the tag-source step for Git.

    Note
    ----
    This makes extensive use of the python sh library. This was a deliberate choice,
    as the gitpython library doesn't appear to easily support username/password auth
    for http and https git repos, and that is a desired use case.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """
        Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return []

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * either both git-username and git-password are set or neither.

        Raises
        ------
        StepRunnerException
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        # ensure either both git-username and git-password are set or neither
        runtime_auth_config = {}
        for auth_config_key in AUTHENTICATION_CONFIG:
            runtime_auth_config_value = self.get_value(auth_config_key)

            if runtime_auth_config_value is not None:
                runtime_auth_config[auth_config_key] = runtime_auth_config_value

        if (any(element in runtime_auth_config for element in AUTHENTICATION_CONFIG)) and \
                (not all(element in runtime_auth_config for element in AUTHENTICATION_CONFIG)):
            raise StepRunnerException(
                "Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            )

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        username = None
        password = None

        if self.has_config_value(AUTHENTICATION_CONFIG):
            username = self.get_value('git-username')
            password = self.get_value('git-password')
        else:
            print('No username/password found, assuming ssh')

        tag = self.__get_tag()

        try:
            self.__git_tag(tag)
            git_url = self.__git_url()
            if git_url.startswith('http://'):
                if username and password:
                    self.__git_push('http://' + username + ':' + password + '@' + git_url[7:])
                else:
                    step_result.success = False
                    step_result.message = 'For a http:// git url, you need to also provide ' \
                                        'username/password pair'
                    return step_result
            elif git_url.startswith('https://'):
                if username and password:
                    self.__git_push('https://' + username + ':' + password + '@' + git_url[8:])
                else:
                    step_result.success = False
                    step_result.message = 'For a https:// git url, you need to also provide ' \
                                        'username/password pair'
                    return step_result
            else:
                self.__git_push(None)
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = str(error)

        step_result.add_artifact(
            name='tag',
            value=tag
        )

        return step_result

    def __get_tag(self):
        tag = self.get_value('version')
        if tag is None:
            tag = 'latest'
            print('No version found in metadata. Using latest')
        return tag

    def __git_url(self):
        git_url = None
        if self.get_value('url'):
            git_url = self.get_value('url')
        else:
            try:
                out = StringIO()
                sh.git.config(
                    '--get',
                    'remote.origin.url',
                    _encoding='UTF-8',
                    _decode_errors='ignore',
                    _out=out,
                    _err=sys.stderr,
                    _tee='err'
                )
                git_url = out.getvalue().rstrip()

                # remove ANYTHING@ from beginning of git_url since step will pass in its own
                # username and password
                #
                # Regex:
                #   ^[^@]+@ - match from beginning of line any character up until
                #             an @ and then the @
                #   (.*) - match any character and capture to capture group 1
                #   \1 - capture group 1 which is the http or https if there was one
                #   \2 - capture group 2 which is anything after the first @ if there was one
                git_url = re.sub(r"^(http://|https://)[^@]+@(.*)", r"\1\2", git_url)

            except sh.ErrorReturnCode as error:  # pylint: disable=undefined-variable
                raise StepRunnerException(
                    f"Error invoking git config --get remote.origin.url: {error}"
                ) from error
        return git_url

    @staticmethod
    def __git_tag(git_tag_value):
        try:
            # NOTE:
            # this force is only needed locally in case of a re-run of the same pipeline
            # without a fresh check out. You will notice there is no force on the push
            # making this an acceptable work around to the issue since on the off chance
            # actually overwriting a tag with a different comment, the push will fail
            # because the tag will be attached to a different git hash.
            sh.git.tag(  # pylint: disable=no-member
                git_tag_value,
                '-f',
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )
        except sh.ErrorReturnCode as error:  # pylint: disable=undefined-variable
            raise StepRunnerException(
                f"Error pushing git tag ({git_tag_value}): {error}"
            ) from error

    @staticmethod
    def __git_push(url=None):
        try:
            if url:
                sh.git.push(
                    url,
                    '--tag',
                    _out=sys.stdout,
                    _err=sys.stderr,
                    _tee='err'
                )
            else:
                sh.git.push(
                    '--tag',
                    _out=sys.stdout,
                    _err=sys.stderr,
                    _tee='err'
                )
        except sh.ErrorReturnCode as error:  # pylint: disable=undefined-variable
            raise StepRunnerException(f"Error invoking git push: {error}") from error



"""`StepImplementers` for the `validate-environment-configuration` step.
"""
from ploigos_step_runner.step_implementers.validate_environment_configuration.\
    configlint import Configlint
from ploigos_step_runner.step_implementers.validate_environment_configuration.\
    configlint_from_argocd import ConfiglintFromArgocd





from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.config.config_value_decryptor import ConfigValueDecryptor
from ploigos_step_runner.decryption_utils import DecryptionUtils
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.io import TextIOSelectiveObfuscator
from ploigos_step_runner.config.decryptors.sops import SOPS

from contextlib import redirect_stdout
import io
import unittest
import re
import sys

from tests.helpers.base_test_case import BaseTestCase

class SampleConfigValueDecryptor(ConfigValueDecryptor):
    ENCRYPTED_VALUE_REGEX = r'^TEST_ENC\[(.*)\]$'

    def can_decrypt(self, config_value):
        return re.match(
            SampleConfigValueDecryptor.ENCRYPTED_VALUE_REGEX,
            config_value.raw_value
        ) is not None

    def decrypt(self, config_value):
        return re.match(
            SampleConfigValueDecryptor.ENCRYPTED_VALUE_REGEX,
            config_value.raw_value
        ).group(1)

class RequiredConstructorParamsConfigValueDecryptor(ConfigValueDecryptor):
    ENCRYPTED_VALUE_REGEX = r'^TEST_ENC\[(.*)\]$'

    def __init__(self, required_arg):
        assert required_arg is not None

    def can_decrypt(self, config_value):
        return re.match(
            SampleConfigValueDecryptor.ENCRYPTED_VALUE_REGEX,
            config_value.raw_value
        ) is not None

    def decrypt(self, config_value):
        return re.match(
            SampleConfigValueDecryptor.ENCRYPTED_VALUE_REGEX,
            config_value.raw_value
        ).group(1)

class BadConfigValueDecryptor:
    pass

class TestDecryptionUtils(BaseTestCase):
    def test_decrypt_no_decryptors(self):
        config_value = ConfigValue(
            'attempt to decrypt me'
        )
        decrypted_value = DecryptionUtils.decrypt(
            config_value
        )
        self.assertIsNone(decrypted_value)

    def test_decrypt_sample_decryptor_does_not_match(self):
        config_value = ConfigValue(
            'attempt to decrypt me'
        )

        DecryptionUtils.register_config_value_decryptor(
            SampleConfigValueDecryptor()
        )

        decrypted_value = DecryptionUtils.decrypt(
            config_value
        )
        self.assertIsNone(decrypted_value)

    def test_decrypt_sample_decryptor(self):
        secret_value = "decrypt me"
        config_value = ConfigValue(
            f'TEST_ENC[{secret_value}]'
        )

        DecryptionUtils.register_config_value_decryptor(
            SampleConfigValueDecryptor()
        )

        decrypted_value = DecryptionUtils.decrypt(
            config_value
        )
        self.assertEqual(
            decrypted_value,
            secret_value
        )

    def test_register_obfuscation_stream(self):
        secret_value = "decrypt me"
        config_value = ConfigValue(
            f'TEST_ENC[{secret_value}]'
        )

        DecryptionUtils.register_config_value_decryptor(
            SampleConfigValueDecryptor()
        )

        out = io.StringIO()
        with redirect_stdout(out):
            old_stdout = sys.stdout
            new_stdout = TextIOSelectiveObfuscator(old_stdout)
            DecryptionUtils.register_obfuscation_stream(new_stdout)

            try:
                sys.stdout = new_stdout
                DecryptionUtils.decrypt(config_value)

                print(f"ensure that I can't actually leak secret value ({secret_value})")
                self.assertRegex(
                    out.getvalue(),
                    r"ensure that I can't actually leak secret value \(\*+\)"
                )
            finally:
                new_stdout.close()
                sys.stdout = old_stdout

    def test__get_decryption_class_sops_short_name(self):
        decryptor_class = DecryptionUtils._DecryptionUtils__get_decryption_class('SOPS')
        self.assertEqual(
            decryptor_class,
            SOPS
        )

    def test__get_decryption_class_sops_full_name(self):
        decryptor_class = DecryptionUtils._DecryptionUtils__get_decryption_class(
            'ploigos_step_runner.config.decryptors.sops.SOPS'
        )
        self.assertEqual(
            decryptor_class,
            SOPS
        )

    def test__get_decryption_class_does_not_exist_short_name(self):
        with self.assertRaisesRegex(
            StepRunnerException,
            r"Could not dynamically load decryptor implementer \(DoesNotExist\) " \
            r"from module \(ploigos_step_runner.config.decryptors\) with class name \(DoesNotExist\)"
        ):
            DecryptionUtils._DecryptionUtils__get_decryption_class('DoesNotExist')

    def test__get_decryption_class_does_not_exist_with_module_name(self):
        with self.assertRaisesRegex(
            StepRunnerException,
            r"Could not dynamically load decryptor implementer \(foo.bar.hello.DoesNotExist\) " \
            r"from module \(foo.bar.hello\) with class name \(DoesNotExist\)"
        ):
            DecryptionUtils._DecryptionUtils__get_decryption_class('foo.bar.hello.DoesNotExist')

    def test__get_decryption_class_not_correct_type(self):
        with self.assertRaisesRegex(
            StepRunnerException,
            r"For decryptor implementer \(tests.test_decryption_utils.BadConfigValueDecryptor\) " \
            r"dynamically loaded class \(<class 'tests.test_decryption_utils." \
            r"BadConfigValueDecryptor'>\) which is not sub class of " \
            r"\(<class 'ploigos_step_runner.config.config_value_decryptor.ConfigValueDecryptor'>\) and should be."
        ):
            DecryptionUtils._DecryptionUtils__get_decryption_class(
                'tests.test_decryption_utils.BadConfigValueDecryptor'
            )

    def test_create_and_register_config_value_decryptor_no_constructor_args(self):
        DecryptionUtils.create_and_register_config_value_decryptor(
            'tests.test_decryption_utils.SampleConfigValueDecryptor'
        )

        secret_value = "decrypt me"
        config_value = ConfigValue(
            f'TEST_ENC[{secret_value}]'
        )
        decrypted_value = DecryptionUtils.decrypt(
            config_value
        )
        self.assertEqual(
            decrypted_value,
            secret_value
        )

    def test_create_and_register_config_value_decryptor_required_constructor_args_missing_arg(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Loaded decryptor class \(<class 'tests.test_decryption_utils." +
            r"RequiredConstructorParamsConfigValueDecryptor'>\) failed to construct with " +
            r"given constructor arguments \({}\): __init__\(\) missing 1 " +
            r"required positional argument: 'required_arg'"
        ):
            DecryptionUtils.create_and_register_config_value_decryptor(
                'tests.test_decryption_utils.RequiredConstructorParamsConfigValueDecryptor'
            )

    def test_create_and_register_config_value_decryptor_required_constructor_args(self):
        DecryptionUtils.create_and_register_config_value_decryptor(
            'tests.test_decryption_utils.RequiredConstructorParamsConfigValueDecryptor',
            {
                'required_arg': 'hello world'
            }
        )

        secret_value = "decrypt me"
        config_value = ConfigValue(
            f'TEST_ENC[{secret_value}]'
        )
        decrypted_value = DecryptionUtils.decrypt(
            config_value
        )
        self.assertEqual(
            decrypted_value,
            secret_value
        )



"""`StepImplementers` for the `package` step.
"""

from ploigos_step_runner.step_implementers.package.maven_package import MavenPackage
from ploigos_step_runner.step_implementers.package.maven import Maven
from ploigos_step_runner.step_implementers.package.npm_package import NpmPackage



"""Shared utils for dealing with containers.
"""

import sys
from io import StringIO

import sh
from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback


def container_registries_login(  #pylint: disable=too-many-branches
    registries,
    containers_config_auth_file=None,
    containers_config_tls_verify=True,
    container_command_short_name=None
):
    """Logs into one or more container registries.

    Requires one of the following to be installed to do the authentication:
    * buidlah
    * podman
    * skopeo

    Notes
    -----
    registries example 1 (dict of dicts where child dict keys are registry uri):

        {
            'registry.redhat.io': {
                'username': 'hello@world.xyz',
                'password': 'nope'
            },
            'registry.internal.example.xyz': {
                'username': 'hello@example.xyz',
                'password': 'nope'
            }
        }

    registries example 2 (dict of dicts where uri is key in child dicts):

        {
            'redhat': {
                'uri': registry.redhat.io
                'username': 'hello@world.xyz',
                'password': 'nope'
            },
            'internal': {
                'uri': 'registry.internal.example.xyz'
                'username': 'hello@example.xyz',
                'password': 'nope'
            }
        }

    registries example 3 (list of dicts where uri is key in child dicts):

        [
            {
                'uri': registry.redhat.io
                'username': 'hello@world.xyz',
                'password': 'nope'
            },
            {
                'uri': 'registry.internal.example.xyz'
                'username': 'hello@example.xyz',
                'password': 'nope'
            }
        ]

    Parameters
    ----------
    registries : dict or list or None
        Dict of dicts of registry configurations or a list of dicts of registry configurations.
        See Notes section for details.
        If none, does nothing.
    containers_config_auth_file : str, optional
        Path of the authentication file.
        If not specified default of the underlying authentication system will be used.
    container_command_short_name : str, optional
        Short name for the command to log in with.
        If not provided will pick the first command found in order (buildah, podman, skopeo).

    See Also
    --------
    container_registry_login : Performs the login for a single container registry
    sh.buildah : https://www.mankier.com/1/buildah-login
    sh.podman : https://www.mankier.com/1/podman-login
    sh.skopeo : https://www.mankier.com/1/skopeo-login
    """

    if registries is None:
        return

    assert isinstance(registries, (dict, list))

    if isinstance(registries, dict):
        for registry_key, registry_conf in registries.items():
            if isinstance(registry_conf, ConfigValue):
                registry_conf = registry_conf.value

            assert 'username' in registry_conf, \
                f"Configuration for container registry ({registry_key}) " \
                f"must specify a 'username': {registry_conf}"
            assert 'password' in registry_conf, \
                f"Configuration for container registry ({registry_key}) " \
                f"must specify a 'password': {registry_conf}"

            # if uri is key in the dict then use that for the registry uri
            # else use the dict key for the registry conf as the registry uri
            if 'uri' in registry_conf:
                registry_uri = registry_conf['uri']
            else:
                registry_uri = registry_key

            if containers_config_tls_verify:
                if 'tls-verify' in registry_conf:
                    registry_tls_verify = registry_conf['tls-verify']
                else:
                    registry_tls_verify = True
            else:
                registry_tls_verify = False

            container_registry_login(
                container_registry_uri=registry_uri,
                container_registry_username=registry_conf['username'],
                container_registry_password=registry_conf['password'],
                container_registry_tls_verify=registry_tls_verify,
                containers_config_auth_file=containers_config_auth_file,
                container_command_short_name=container_command_short_name
            )
    elif isinstance(registries, list):
        for registry_conf in registries:
            if isinstance(registry_conf, ConfigValue):
                registry_conf = registry_conf.value

            assert 'uri' in registry_conf, \
                f"Configuration for container registry " \
                f"must specify a 'uri': {registry_conf}"
            assert 'username' in registry_conf, \
                f"Configuration for container registry " \
                f"must specify a 'username': {registry_conf}"
            assert 'password' in registry_conf, \
                f"Configuration for container registry " \
                f"must specify a 'password': {registry_conf}"

            if containers_config_tls_verify:
                if 'tls-verify' in registry_conf:
                    registry_tls_verify = registry_conf['tls-verify']
                else:
                    registry_tls_verify = True
            else:
                registry_tls_verify = False

            container_registry_login(
                container_registry_uri=registry_conf['uri'],
                container_registry_username=registry_conf['username'],
                container_registry_password=registry_conf['password'],
                container_registry_tls_verify=registry_tls_verify,
                containers_config_auth_file=containers_config_auth_file,
                container_command_short_name=container_command_short_name
            )

def container_registry_login( #pylint: disable=too-many-arguments,too-many-branches
    container_registry_uri,
    container_registry_username,
    container_registry_password,
    container_registry_tls_verify=True,
    containers_config_auth_file=None,
    container_command_short_name=None
):
    """Performs the login for a single container registry.

    Requires one of the following to be installed to do the authentication:
    * buidlah
    * podman
    * skopeo

    Parameters
    ----------
    container_registry_uri : str or ConfigValue
        URI to the container registry to log into.
    container_registry_username : str or ConfigValue
        Username to log into the container registry with.
    container_registry_password : str or ConfigValue
        Password to log into the container registry with.
    container_registry_tls_verify : bool or str or ConfigValue
        True to verify container registry certificates as part of authenticating.
        False to ignore certificate chain.
        NOTE: no matter what SSL is used to authenticate with container registry
    containers_config_auth_file : str or ConfigValue, optional
        Path of the authentication file.
        If not specified default of the underlying authentication system will be used.
    container_command_short_name : str, optional
        Short name for the command to log in with.
        If not provided will pick the first command found in order (buildah, podman, skopeo).

    Raises
    ------
    RuntimeError
        When can not find tool to login to container registry with.
        When error loging into container registry.

    See Also
    --------
    container_registries_login : authenticate with multiple container registries
    sh.buildah : https://www.mankier.com/1/buildah-login
    sh.podman : https://www.mankier.com/1/podman-login
    sh.skopeo : https://www.mankier.com/1/skopeo-login
    """

    assert container_registry_uri
    assert container_registry_username
    assert container_registry_password

    if isinstance(container_registry_uri, ConfigValue):
        container_registry_uri = container_registry_uri.value
    if isinstance(container_registry_username, ConfigValue):
        container_registry_username = container_registry_username.value
    if isinstance(container_registry_password, ConfigValue):
        container_registry_password = container_registry_password.value
    if isinstance(container_registry_tls_verify, ConfigValue):
        container_registry_tls_verify = container_registry_tls_verify.value
    if isinstance(containers_config_auth_file, ConfigValue):
        containers_config_auth_file = containers_config_auth_file.value

    # can use any of these tools to authenticate, look for them all and use first available
    #
    # NOTE: this all works because these three commands take the exact same parameters for login
    # if implementing some new command, like docker, you will need to deal with the differences
    buildah_path = sh.which('buildah')
    podman_path = sh.which('podman')
    skopeo_path = sh.which('skopeo')
    if container_command_short_name:
        given_command_path = sh.which(container_command_short_name)
        if given_command_path:
            container_command = sh.Command(container_command_short_name).bake()
        else:
            raise RuntimeError(
                f"When attempting to login to container registry ({container_registry_uri}) "
                f"could not find the given expected tool ({container_command_short_name}) "
                "to login with."
            )
    elif buildah_path is not None:
        container_command = sh.buildah.bake() #pylint: disable=no-member
    elif podman_path is not None:
        container_command = sh.podman.bake() #pylint: disable=no-member
    elif skopeo_path is not None:
        container_command = sh.skopeo.bake() #pylint: disable=no-member
    else:
        raise RuntimeError(
            f"When attempting to login to container registry ({container_registry_uri}) "
            "could not find one of the expected tools (buildah, podman, skopeo) to login with."
        )

    login_command_named_flags = {
        'password_stdin': True,
        'username': container_registry_username,
        'tls_verify': str(container_registry_tls_verify).lower()
    }
    if containers_config_auth_file:
        login_command_named_flags['authfile'] = containers_config_auth_file

    try:
        # NOTE: need to bake in the flags prior to the registry due to nonesense with
        #       required ordering by the login command and how sh handles
        #       escaping and ordering parameters
        print(f"Login ({container_command}) to container image registry ({container_registry_uri})")
        login_comnmand = container_command.login.bake(**login_command_named_flags)
        login_comnmand(
            container_registry_uri,
            _in=container_registry_password,
            _out=sys.stdout,
            _err=sys.stderr,
            _tee='err'
        )
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f"Failed to login to container registry ({container_registry_uri}) "
            f"with username ({container_registry_username}): {error}"
        ) from error

def create_container_from_image(
    image_tag,
    repository_type='container-storage:'
):
    """Import a container image using buildah form a TAR file.

    Parameters
    ----------
    image_tag : str
        Image tag to create a container from.
        ex:
        * localhost/my-app:latest
        * quay.io/my-org/my-app:latest
        * docker-archive:/local/path/to/my-app-container-image.tar
    container_name : str
        name for the working container.
    repository_type : str
        The type of repository to mount the given image tag from.
        See https://github.com/containers/skopeo for details on different repository types.

    Returns
    -------
    str
        Name of the imported container.

    Raises
    ------
    RuntimeError
        If error importing image.
    """
    container_name = None
    try:
        buildah_from_out_buff = StringIO()
        buildah_from_out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
            sys.stdout,
            buildah_from_out_buff
        ])
        sh.buildah(  # pylint: disable=no-member
            'from',
            f"{repository_type}{image_tag}",
            _out=buildah_from_out_callback,
            _err=sys.stderr,
            _tee='err'
        )
        container_name = buildah_from_out_buff.getvalue().rstrip()
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f'Error creating container from image ({image_tag}): {error}'
        ) from error

    return container_name


def mount_container(buildah_unshare_command, container_id):
    """Use buildah to mount a container.

    Parameters
    ----------
    buildah_unshare_command : sh.buildah.unshare.bake()
        A baked sh.buildah.unshare command to use to run this command in the context off
        so that this can be done "rootless".
    container_id : str
        ID of the container to mount.

    Returns
    -------
    str
        Absolute path to the mounted container.

    Raises
    ------
    RuntimeError
        If error mounting the container.
    """
    mount_path = None
    try:
        buildah_mount_out_buff = StringIO()
        buildah_mount_out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
            sys.stdout,
            buildah_mount_out_buff
        ])
        buildah_mount_command = buildah_unshare_command.bake("buildah", "mount")
        buildah_mount_command(
            container_id,
            _out=buildah_mount_out_callback,
            _err=sys.stderr,
            _tee='err'
        )
        mount_path = buildah_mount_out_buff.getvalue().rstrip()
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f'Error mounting container ({container_id}): {error}'
        ) from error

    return mount_path

def determine_container_image_build_tag_info(
    image_version,
    organization,
    application_name,
    service_name
):
    """Determines the full and short build tags for a new container image.

    Parameters
    ----------
    image_version : str
        A given image version. If none given, latest will be used.
    organization : str
        Organization the container image belongs to.
    application_name : str
        Application the container image belongs to.
    service_name : str
        Service the container image implements.

    Returns
    -------
    str, str, str, str, str
        First result is the full build tag, including registry URI.
        Second result is the short build tag, as in no registry URI.
        Third result is the image registry uri.
        Forth result is the image repository name.
        Fifth result is the used image version.

    """
    if image_version is None:
        image_version = 'latest'
        print('No image tag version found in metadata. Using latest')
    image_registry_uri = 'localhost'
    image_registry_organization = organization
    image_repository = f"{application_name}-{service_name}"
    build_short_tag = f"{image_registry_organization}/{image_repository}:{image_version}"
    build_full_tag = f"{image_registry_uri}/{build_short_tag}"

    return build_full_tag, build_short_tag, image_registry_uri, image_repository, image_version



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os

from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase

from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.generate_metadata import SemanticVersion


class TestStepImplementerSemanticVersionGenerateMetadata(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=SemanticVersion,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = SemanticVersion.step_implementer_config_defaults()
        expected_defaults = {
            'release-branch': 'master'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = SemanticVersion._required_config_or_result_keys()
        expected_required_keys = [
            'app-version',
            'pre-release',
            'release-branch',
            'build'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    def test_run_step_pass(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {}

            artifact_config = {
                'app-version': {'description': '', 'value': '42.1.0'},
                'pre-release': {'description': '', 'value': 'master'},
                'build': {'description': '', 'value': 'abc123'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='SemanticVersion',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='SemanticVersion',
                sub_step_implementer_name='SemanticVersion'
            )
            expected_step_result.add_artifact(name='version', value='42.1.0+abc123')
            expected_step_result.add_artifact(name='container-image-version', value='42.1.0')

            expected_step_result.add_evidence(name='version', value='42.1.0+abc123')
            expected_step_result.add_evidence(name='container-image-version', value='42.1.0')

            self.assertEqual(result, expected_step_result)



    def test_run_step_pass_different_pre_release(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_config = {}

            artifact_config = {
                'app-version': {'description': '', 'value': '42.1.0'},
                'pre-release': {'description': '', 'value': 'feature123'},
                'build': {'description': '', 'value': 'abc123'}
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='SemanticVersion',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='SemanticVersion',
                sub_step_implementer_name='SemanticVersion'
            )
            expected_step_result.add_artifact(name='version', value='42.1.0-feature123+abc123')
            expected_step_result.add_artifact(name='container-image-version', value='42.1.0-feature123')
            
            expected_step_result.add_evidence(name='version', value='42.1.0-feature123+abc123')
            expected_step_result.add_evidence(name='container-image-version', value='42.1.0-feature123')

            self.assertEqual(result, expected_step_result)



"""`StepImplementer` for the `create-container-image step` using Buildah.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key  | Required? | Default | Description
-------------------|-----------|---------|-----------
`imagespecfile`    | True      | `'Containerfile'` \
                                         | File defining the container image
`context`          | True      | `'.'`   | Context to build the container image in
`tls-verify`       | True      | `True`  | Whether to verify TLS when pulling parent images
`format`           | True      | `'oci'` | format of the built image's manifest and metadata
`containers-config-auth-file` \
                   | False     |         | Path to the container registry authentication \
                                           file to use for container registry authentication. \
                                           If one is not provided one will be created in the \
                                           working directory.
`container-image-version` \
                   | True      |         | Version to use when building the container image
`organization`     | True      |         | Used in built container image tag
`application_name` | True      |         | Used in built container image tag
`service_name`     | True      |         | Used in built container image tag
`container-registries` \
                   | False     |         | Hash of container registries to authenticate with.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key                     | Description
----------------------------------------|------------
`container-image-registry-uri`          | Registry URI poriton of the container image tag of the built container image.
`container-image-registry-organization` | Organization portion of the container image tag of the built container image.
`container-image-repository`            | Repository portion of the container image tag of the built container image.
`container-image-name`                  | Another way to reference the repository portion of the container image tag \
                                          of the built container image.
`container-image-version`               | Version portion of the container image tag of the built container image.
`container-image-tag`                   | Full container image tag of the built container, including the registry URI. <br/> \
                                          Takes the form of: \
                                          `container-image-registry-organization/container-image-repository:container-image-version`
`container-image-short-tag`             | Short container image tag of the built container image,  excluding the registry URI. <br/> \
                                          Takes the form of: \
                                          `container-image-registry-uri/container-image-registry-organization/container-image-repository:container-image-version`

""" # pylint: disable=line-too-long

import os
import sys
from distutils import util

import sh
from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.containers import (
    container_registries_login, determine_container_image_build_tag_info)

DEFAULT_CONFIG = {
    # Image specification file name
    'imagespecfile': 'Containerfile',

    # Parent path to the image specification file
    'context': '.',

    # Verify TLS Certs?
    'tls-verify': True,

    # Format of the produced image
    'format': 'oci'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'imagespecfile',
    'context',
    'tls-verify',
    'format',
    'organization',
    'service-name',
    'application-name'
]

class Buildah(StepImplementer):
    """`StepImplementer` for the `create-container-image step` using Buildah.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * given 'imagespecfile' exists

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        # if pom-file has value verify file exists
        # If it doesn't have value and is required function will have already failed
        image_spec_file = self.get_value('imagespecfile')
        context = self.get_value('context')
        image_spec_file_full_path = os.path.join(context, image_spec_file)
        assert os.path.exists(image_spec_file_full_path), \
            f'Given imagespecfile ({image_spec_file}) does not exist in given context ({context}).'

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # get config
        image_spec_file = self.get_value('imagespecfile')
        tls_verify = self.get_value('tls-verify')
        if isinstance(tls_verify, str):
            tls_verify = bool(util.strtobool(tls_verify))

        # create local build tag
        image_registry_organization = self.get_value('organization')
        build_full_tag, build_short_tag, image_registry_uri, image_repository, image_version = \
            determine_container_image_build_tag_info(
                image_version=self.get_value('container-image-version'),
                organization=image_registry_organization,
                application_name=self.get_value('application-name'),
                service_name=self.get_value('service-name')
            )

        try:
            # login to any provider container registries
            # NOTE: important to specify the auth file because depending on the context this is
            #       being run in python process may not have permissions to default location
            containers_config_auth_file = self.get_value('containers-config-auth-file')
            if not containers_config_auth_file:
                containers_config_auth_file = os.path.join(
                    self.work_dir_path,
                    'container-auth.json'
                )
            container_registries_login(
                registries=self.get_value('container-registries'),
                containers_config_auth_file=containers_config_auth_file,
                containers_config_tls_verify=tls_verify
            )

            # perform build
            sh.buildah.bud(  # pylint: disable=no-member
                '--format=' + self.get_value('format'),
                '--tls-verify=' + str(tls_verify).lower(),
                '--layers', '-f', image_spec_file,
                '-t', build_full_tag,
                '--authfile', containers_config_auth_file,
                self.get_value('context'),
                _out=sys.stdout,
                _err=sys.stderr,
                _tee='err'
            )

        except sh.ErrorReturnCode as error:  # pylint: disable=undefined-variable
            step_result.success = False
            step_result.message = 'Issue invoking buildah bud with given image ' \
                f'specification file ({image_spec_file}): {error}'
            return step_result

        # add artifacts
        step_result.add_artifact(
            name='container-image-registry-uri',
            value=image_registry_uri,
            description='Registry URI poriton of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-registry-organization',
            value=image_registry_organization,
            description='Organization portion of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-repository',
            value=image_repository,
            description='Repository portion of the container image tag' \
                ' of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-name',
            value=image_repository,
            description='Another way to reference the' \
                ' repository portion of the container image tag of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-version',
            value=image_version,
            description='Version portion of the container image tag of the built container image.'
        )
        step_result.add_artifact(
            name='container-image-tag',
            value=build_full_tag,
            description='Full container image tag of the built container,' \
                ' including the registry URI.'
        )
        step_result.add_artifact(
            name='container-image-short-tag',
            value=build_short_tag,
            description='Short container image tag of the built container image,' \
                ' excluding the registry URI.'
        )

        return step_result



"""Representation of configuration for workflow.
"""

import copy
import glob
import os.path
from distutils import util

from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.config.step_config import StepConfig
from ploigos_step_runner.decryption_utils import DecryptionUtils
from ploigos_step_runner.utils.dict import deep_merge
from ploigos_step_runner.utils.file import parse_yaml_or_json_file


class Config:
    """Representation of configuration for Ploigos workflow.

    Parameters
    ----------
    config : dict, list, str (file or directory), optional
        A dictionary that is a valid configuration,
        or a string that is a path to a YAML or JSON file that is
        a valid configuration,
        or a string that is a path to a directory containing one or more
        files that are valid YAML or JSON files that are valid
        configurations,
        or a list of any of the former.

    Attributes
    ----------
    __global_defaults : dict
    __global_environment_defaults : dict
    __step_configs : dict of str (step names) to StepConfig

    Raises
    ------
    ValueError
        If given config is not of expected type.
    AssertionError
        If given config contains any invalid configurations.
    """
    CONFIG_KEY = 'step-runner-config'
    CONFIG_KEY_GLOBAL_DEFAULTS = 'global-defaults'
    CONFIG_KEY_GLOBAL_ENVIRONMENT_DEFAULTS = 'global-environment-defaults'
    CONFIG_KEY_ENVIRONMENT_NAME = 'environment-name'
    CONFIG_KEY_CONTINUE_SUB_STEPS_ON_FAILURE = 'continue-sub-steps-on-failure'
    CONFIG_KEY_STEP_IMPLEMENTER = 'implementer'
    CONFIG_KEY_SUB_STEP_NAME = 'name'
    CONFIG_KEY_SUB_STEP_CONFIG = 'config'
    CONFIG_KEY_SUB_STEP_ENVIRONMENT_CONFIG = 'environment-config'
    CONFIG_KEY_DECRYPTORS = 'config-decryptors'
    CONFIG_KEY_DECRYPTOR_IMPLEMENTER = 'implementer'
    CONFIG_KEY_DECRYPTOR_CONFIG = 'config'

    def __init__(self, config=None):
        self.__global_defaults = {}
        self.__global_environment_defaults = {}
        self.__step_configs = {}

        if config is not None:
            self.add_config(config)

    @property
    def global_defaults(self):
        """Get a deep copy of the global defaults.

        Returns
        -------
        dict
            Deep copy of the global defaults.
        """
        return copy.deepcopy(self.__global_defaults)

    @property
    def global_environment_defaults(self):
        """Deep copy of all global environment defaults for all environments.

        Returns
        -------
        dict
            deep copy of all global environment defaults.
        """
        return copy.deepcopy(self.__global_environment_defaults)

    @property
    def step_configs(self):
        """
        Returns
        -------
        list of StepConfig
        """
        return self.__step_configs

    def get_global_environment_defaults_for_environment(self, env):
        """Get a deep copy of all of the global environment defaults or for a given an environment.

        Parameters
        ----------
        env : str
            The global environment defaults for the given environment.
            If given environment name does not exist in environment defaults then empty dict.

        Returns
        -------
        dict
            Deep copy of the global environment defaults for the given environment
            or empty dict if no environment given or environment does not exist in the defaults
        """
        if env is not None:
            if env in self.__global_environment_defaults:
                global_environment_defaults = copy.deepcopy(self.__global_environment_defaults[env])
            else:
                global_environment_defaults = {}
        else:
            global_environment_defaults = {}

        return global_environment_defaults

    def get_step_config(self, step_name):
        """Get the step config for a given step name.

        Parameters
        ----------
        step_name : str
            Name of the step to get the step configuration for.

        Returns
        -------
        StepConfig
            Step configuration for the given step name or None if does not exist
        """
        if step_name in self.step_configs:
            step_config = self.step_configs[step_name]
        else:
            step_config = None

        return step_config

    def get_sub_step_configs(self, step_name):
        """Gets lit of configured sub step configurations for a step with the given name.

        Parameters
        ----------
        step_name : str
            Name of step to get configured sub steps for.

        Returns
        -------
        list of SubStepConfig
            List of configured sub step configurations for the step with the given name.
        """

        if step_name in self.step_configs:
            sub_step_configs = self.step_configs[step_name].sub_steps
        else:
            sub_step_configs = []

        return sub_step_configs

    def add_config(self, config):
        """Parses, validates, and adds a given config to this Config.

        Parameters
        ----------
        config : dict, list, str (file or directory)
            A dictionary that is a valid configuration,
            or aa string that is a path to a YAML or JSON file that is
            a valid configuration,
            or a string that is a path to a directory containing one or more
            files that are valid YAML or JSON files that are valid
            configurations,
            or a list of any of the former.

        Raises
        ------
        ValueError
            If given config is not of expected type.
        AssertionError
            If given config contains any invalid configurations.
        """
        if isinstance(config, dict):
            # add the config
            self.__add_config_dict(config)
        elif isinstance(config, list):
            # for each item in the list, treat it as a config to add
            for _config in config:
                self.add_config(_config)
        elif isinstance(config, str):
            if os.path.isfile(config):
                self.__add_config_file(config)
            elif os.path.isdir(config):
                # for each recursively found file in the directory add the config file
                config_dir_files = glob.glob(config + '/**', recursive=True)
                found_nested_file = False
                for config_dir_file in config_dir_files:
                    if os.path.isfile(config_dir_file):
                        found_nested_file = True
                        self.__add_config_file(config_dir_file)

                if not found_nested_file:
                    raise ValueError(
                        f"Given config string ({config}) is a directory" +
                        " with no recursive children files."
                    )
            else:
                raise ValueError(
                    f"Given config string ({config}) is not a valid path."
                )
        else:
            raise ValueError(
                f"Given config ({config}) is unexpected type ({type(config)}) " +
                "not a dictionary, string, or list of former."
            )

    def set_step_config_overrides(self, step_name, step_config_overrides):
        """Sets configuration overrides for all sub steps of a given step.

        Notes
        -----
        If step configuration overrides are already set for the given step then
        the existing overides will be replaced with the newly given step configuration
        overrides.

        Parameters
        ----------
        step_name : str
            Name of step to add configuration overrides for all sub steps for.
        step_config_overrides : dict
            Overrides for all sub steps for the step with the given name.
        """
        if step_name not in self.step_configs:
            self.step_configs[step_name] = StepConfig(self, step_name)

        self.step_configs[step_name].step_config_overrides = step_config_overrides

    def __add_config_file(self, config_file):
        """Adds a JSON or YAML file as config to this Config.

        Parameters
        ----------
        config_file : str (file path)
            A string that is a path to an existing YAML or JSON file to
            parse and validate as a configuration to add to this Config.

        Raises
        ------
        ValueError
            If can not parse given file as YAML or JSON
        AssertionError
            If dictionary parsed from given YAML or JSON file is not a valid config.
        """
        # parse the configuration file
        try:
            parsed_config_file = parse_yaml_or_json_file(config_file)
        except ValueError as error:
            raise ValueError(
                f"Error parsing config file ({config_file}) as json or yaml"
            ) from error

        # add the config parsed from file
        try:
            self.__add_config_dict(parsed_config_file, config_file)
        except AssertionError as error:
            raise AssertionError(
                f"Failed to add parsed configuration file ({config_file}): {error}"
            ) from error

    def __add_config_dict(self, config_dict, source_file_path=None): # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        """Add a configuration dictionary to the list of configuration dictionaries.

        Parameters
        ----------
        config_dict : dict
            A dictionary to validate as a configuration and to add to this Config.
        source_file_path : str, optional
            File path to the file from which the given config_dict came from.

        Raises
        ------
        AssertionError
            If the given config_dict is not a valid configuration dictionary.
            If attempt to update an existing sub step and new and existing sub step implementers
                do not match.
            If sub step does not define a step implementer.
        ValueError
            If duplicative leaf keys when merging global defaults
            If duplicative leaf keys when merging global env defaults
            If step config is not of type dict or list
            If new sub step configuration has duplicative leaf keys to
                existing sub step configuration.
            If new sub step environment configuration has duplicative leaf keys to
                existing sub step environment configuration.
        """

        assert Config.CONFIG_KEY in config_dict, \
            "Failed to add invalid config. " + \
            f"Missing expected top level key ({Config.CONFIG_KEY}): " + \
            f"{config_dict}"

        # if file path given use that as the source when creating ConfigValue objects
        # else use a copy of the given configuration dictionary
        if source_file_path is not None:
            parent_source = source_file_path
        else:
            parent_source = copy.deepcopy(config_dict)

        # convert all the leaves of the configuration dictionary under
        # the Config.CONFIG_KEY to ConfigValue objects
        config_values = ConfigValue.convert_leaves_to_config_values(
            values=copy.deepcopy(config_dict[Config.CONFIG_KEY]),
            parent_source=parent_source,
            path_parts=[Config.CONFIG_KEY]
        )

        for key, value in config_values.items():
            # if global default key
            # else if global env defaults key
            # else assume step config
            if key == Config.CONFIG_KEY_GLOBAL_DEFAULTS:
                try:
                    self.__global_defaults = deep_merge(
                        copy.deepcopy(self.__global_defaults),
                        copy.deepcopy(value)
                    )
                except ValueError as error:
                    raise ValueError(
                        f"Error merging global defaults: {error}"
                    ) from error
            elif key == Config.CONFIG_KEY_GLOBAL_ENVIRONMENT_DEFAULTS:
                for env, env_config in value.items():
                    if env not in self.__global_environment_defaults:
                        self.__global_environment_defaults[env] = {
                            Config.CONFIG_KEY_ENVIRONMENT_NAME: env
                        }

                    try:
                        self.__global_environment_defaults[env] = deep_merge(
                            copy.deepcopy(self.__global_environment_defaults[env]),
                            copy.deepcopy(env_config)
                        )
                    except ValueError as error:
                        raise ValueError(
                            f"Error merging global environment ({env}) defaults: {error}"
                        ) from error
            elif key == Config.CONFIG_KEY_DECRYPTORS:
                config_decryptor_definitions = ConfigValue.convert_leaves_to_values(value)
                Config.parse_and_register_decryptors_definitions(config_decryptor_definitions)
            else:
                step_name = key
                step_config = value

                # if step_config is dict then assume step with single sub step
                if isinstance(step_config, dict):
                    sub_steps = [step_config]
                elif isinstance(step_config, list):
                    sub_steps = step_config
                else:
                    raise ValueError(
                        f"Expected step ({step_name}) to have have step config ({step_config})" +
                        f" of type dict or list but got: {type(step_config)}"
                    )

                for sub_step in sub_steps:
                    assert Config.CONFIG_KEY_STEP_IMPLEMENTER in sub_step, \
                        f"Step ({step_name}) defines a single sub step with values " + \
                        f"({sub_step}) but is missing value for key: " + \
                        f"{Config.CONFIG_KEY_STEP_IMPLEMENTER}"

                    sub_step_implementer_name = \
                        sub_step[Config.CONFIG_KEY_STEP_IMPLEMENTER].value

                    # if sub step name given
                    # else if no sub step name given use step implementer as sub step name
                    if Config.CONFIG_KEY_SUB_STEP_NAME in sub_step:
                        sub_step_name = sub_step[Config.CONFIG_KEY_SUB_STEP_NAME].value
                    else:
                        sub_step_name = sub_step_implementer_name

                    # determine sub step config
                    if Config.CONFIG_KEY_SUB_STEP_CONFIG in sub_step:
                        sub_step_config_dict = copy.deepcopy(
                            sub_step[Config.CONFIG_KEY_SUB_STEP_CONFIG])
                    else:
                        sub_step_config_dict = {}

                    # determine sub step environment config
                    if Config.CONFIG_KEY_SUB_STEP_ENVIRONMENT_CONFIG in sub_step:
                        sub_step_env_config = copy.deepcopy(
                            sub_step[Config.CONFIG_KEY_SUB_STEP_ENVIRONMENT_CONFIG])
                    else:
                        sub_step_env_config = {}

                    # determine if continue sub steps on this sub step failure
                    sub_step_contine_sub_steps_on_failure = False
                    if Config.CONFIG_KEY_CONTINUE_SUB_STEPS_ON_FAILURE in sub_step:
                        sub_step_contine_sub_steps_on_failure = sub_step[
                            Config.CONFIG_KEY_CONTINUE_SUB_STEPS_ON_FAILURE
                        ]
                        if isinstance(sub_step_contine_sub_steps_on_failure.value, bool):
                            sub_step_contine_sub_steps_on_failure = \
                                sub_step_contine_sub_steps_on_failure.value
                        else:
                            sub_step_contine_sub_steps_on_failure = bool(
                                util.strtobool(sub_step_contine_sub_steps_on_failure.value)
                            )

                    self.add_or_update_step_config(
                        step_name=step_name,
                        sub_step_name=sub_step_name,
                        sub_step_implementer_name=sub_step_implementer_name,
                        sub_step_config_dict=sub_step_config_dict,
                        sub_step_env_config=sub_step_env_config,
                        sub_step_contine_sub_steps_on_failure=sub_step_contine_sub_steps_on_failure
                    )

    @staticmethod
    def parse_and_register_decryptors_definitions(decryptors_definitions):
        """Parse decryptor definitions from a list and then register them with the DecryptionUtils.

        Parameters
        ----------
        decryptors_definitions : list of dicts
            List of decryptor definitions. Each element should be a dict with at least an
            'implementer' key with a string value and optionally a 'config' key with a dict value.

        Raises
        ------
        AssertionError
            If decryptors_definitions is not a list.
            If a decryptor definition does not have a
                Config.CONFIG_KEY_DECRYPTOR_IMPLEMENTER key.
        """
        assert isinstance(decryptors_definitions, list), \
            f"Decryptors configuration ({decryptors_definitions}) must be of type " + \
            f"(list) got: {type(decryptors_definitions)}"

        for decryptor_definition in decryptors_definitions:
            assert Config.CONFIG_KEY_DECRYPTOR_IMPLEMENTER in decryptor_definition, \
                "Decryptor configuration is missing key " + \
                f"({Config.CONFIG_KEY_DECRYPTOR_IMPLEMENTER}): {decryptor_definition}"

            decryptor_implementer_name = \
                decryptor_definition[Config.CONFIG_KEY_DECRYPTOR_IMPLEMENTER]

            if Config.CONFIG_KEY_DECRYPTOR_CONFIG in decryptor_definition:
                decryptor_config = decryptor_definition[Config.CONFIG_KEY_DECRYPTOR_CONFIG]
            else:
                decryptor_config = {}

            DecryptionUtils.create_and_register_config_value_decryptor(
                decryptor_implementer_name,
                decryptor_config
            )

    def add_or_update_step_config( # pylint: disable=too-many-arguments
        self,
        step_name,
        sub_step_name,
        sub_step_implementer_name,
        sub_step_config_dict,
        sub_step_env_config,
        sub_step_contine_sub_steps_on_failure=False
    ):
        """Adds a new step configuration with a single new sub step or
        updates an existing step with new or updated sub step.

        Parameters
        ----------
        step_name : str
            Name of step to create or update.
        sub_step_name : str
            Name of the sub step to add or update on the new or updated step.
        sub_step_implementer_name : str
            Name of the sub step implementer for the sub step being added or updated.
            If updating this can not be different then existing sub step with the same name.
        sub_step_config_dict : dict, optional
            Sub step configuration to add or update for named sub step on the new or updated step.
            If updating this can not have any duplicative leaf keys to the existing
                sub step configuration.
        sub_step_env_config : dict, optional
            Sub step environment configuration to add or update for named sub step on the
                new or updated step.
            If updating this can not have any duplicative leaf keys to the existing
                sub step environment configuration.
        sub_step_contine_sub_steps_on_failure : bool
            True to continue executing other sub steps in current step if this sub step fails.
            False to fail all step execution if this sub step fails.

        Raises
        ------
        AssertionError
            If attempt to update an existing sub step and new and existing sub step implementers
                do not match.
        ValueError
            If new sub step configuration has duplicative leaf keys to
                existing sub step configuration.
            If new sub step environment configuration has duplicative leaf keys to
                existing sub step environment configuration.
        """

        if step_name not in self.step_configs:
            self.step_configs[step_name] = StepConfig(self, step_name)

        step_config = self.step_configs[step_name]
        step_config.add_or_update_sub_step_config(
            sub_step_name=sub_step_name,
            sub_step_implementer_name=sub_step_implementer_name,
            sub_step_config_dict=sub_step_config_dict,
            sub_step_env_config=sub_step_env_config,
            sub_step_contine_sub_steps_on_failure=sub_step_contine_sub_steps_on_failure
        )



"""`StepImplementers` for the `generate-metadata` step.
"""

from ploigos_step_runner.step_implementers.generate_metadata.commitizen import Commitizen
from ploigos_step_runner.step_implementers.generate_metadata.git import Git
from ploigos_step_runner.step_implementers.generate_metadata.maven import Maven
from ploigos_step_runner.step_implementers.generate_metadata.npm import Npm
from ploigos_step_runner.step_implementers.generate_metadata.semantic_version import SemanticVersion



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os
import re
from io import IOBase

import yaml
from ploigos_step_runner import StepRunner


def create_sh_side_effect(
        mock_stdout=None,
        mock_stderr=None,
        exception=None
):
    def sh_side_effect(*args, **kwargs):
        if mock_stdout:
            if '_out' in kwargs:
                if callable(kwargs['_out']):
                    kwargs['_out'](mock_stdout)
                elif isinstance(kwargs['_out'], IOBase):
                    kwargs['_out'].write(mock_stdout)
            else:
                return mock_stdout

        if mock_stderr:
            if callable(kwargs['_err']):
                kwargs['_err'](mock_stderr)
            elif isinstance(kwargs['_err'], IOBase):
                kwargs['_err'].write(mock_stderr)

        if exception:
            raise exception

    return sh_side_effect

def create_git_commit_with_sample_file(temp_dir, git_repo,
                                       sample_file_name='sample-file',
                                       commit_message='test'):
    sample_file = os.path.join(temp_dir.path, sample_file_name)
    open(sample_file, 'wb').close()
    git_repo.index.add([sample_file])
    git_repo.index.commit(commit_message)


def Any(cls):
    """
    Source
    ------
    https://stackoverflow.com/questions/21611559/assert-that-a-method-was-called-with-one-argument-out-of-several
    """

    class Any(cls):
        def __init__(self):
            pass

        def __eq__(self, other):
            return True

        def __hash__(self):
            return hash(tuple(self))

        def __repr__(self):
            return 'any'

    return Any()


class StringRegexParam():
    def __init__(self, regex):
        self.__regex = regex

    def __eq__(self, other):
        if isinstance(other, str):
            return re.match(self.__regex, other)
        return False


def create_sops_side_effect(mock_stdout):
    def sops_side_effect(*args, **kwargs):
        kwargs['_out'].write(mock_stdout)

    return sops_side_effect



import os
from pathlib import Path
from unittest.mock import patch

from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult
from ploigos_step_runner.step_implementers.create_container_image import MavenJKubeK8sBuild
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.step_implementers.shared.MavenGeneric.__init__")
class TestStepImplementerMavenJKubeK8sBuild___init__(BaseStepImplementerTestCase):
    def test_defaults(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenJKubeK8sBuild(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None,
            maven_phases_and_goals=['k8s:build']
        )

    def test_given_environment(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenJKubeK8sBuild(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env'
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env',
            maven_phases_and_goals=['k8s:build']
        )

class TestStepImplementerMavenJKubeK8sBuild_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenJKubeK8sBuild.step_implementer_config_defaults(),
            {
                'pom-file': 'pom.xml',
                'tls-verify': True,
                'maven-profiles': [],
                'maven-additional-arguments': [],
                'maven-no-transfer-progress': True,
                'maven-additional-arguments': [
                    '-Dmaven.install.skip=true',
                    '-Dmaven.test.skip=true'
                ]
            }
        )

class TestStepImplementerMavenJKubeK8sBuild__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenJKubeK8sBuild._required_config_or_result_keys(),
            [
                'pom-file'
            ]
        )

@patch('ploigos_step_runner.step_implementers.create_container_image.maven_jkube_k8sbuild.determine_container_image_build_tag_info')
@patch.object(MavenJKubeK8sBuild, '_run_maven_step')
@patch.object(
    MavenJKubeK8sBuild,
    'write_working_file',
    return_value='/mock/mvn_k8s_build_output.txt'
)
class TestStepImplementerMavenJKubeK8sBuild__run_step(
    BaseStepImplementerTestCase
):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenJKubeK8sBuild,
            step_config=step_config,
            step_name='create-container-image',
            implementer='MavenJKubeK8sBuild',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_success_given_version(
        self,
        mock_write_working_file,
        mock_run_maven_step,
        mock_determine_container_image_build_tag_info
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app',
                'container-image-version': '1.0-123abc'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup sideeffects
            artifact_parent_dir = os.path.join(test_dir.path, 'target')
            package_artifact_names = [
                f'my-app.jar'
            ]
            def run_maven_side_effect(mvn_output_file_path, step_implementer_additional_arguments):
                os.makedirs(artifact_parent_dir, exist_ok=True)
                for artifact_name in package_artifact_names:
                    artifact_path = os.path.join(
                        artifact_parent_dir,
                        artifact_name
                    )
                    Path(artifact_path).touch()
            mock_run_maven_step.side_effect = run_maven_side_effect

            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='MavenJKubeK8sBuild',
                sub_step_implementer_name='MavenJKubeK8sBuild'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "create container image.",
                name='maven-jkube-output',
                value='/mock/mvn_k8s_build_output.txt'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='1.0-123abc',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:1.0-123abc',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:1.0-123abc',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_k8s_build_output.txt',
                step_implementer_additional_arguments=[
                    '-Djkube.generator.name=localhost/mock-org/mock-app-mock-service:1.0-123abc'
                ]
            )
            mock_determine_container_image_build_tag_info.assert_called_once_with(
                image_version='1.0-123abc',
                organization='mock-org',
                application_name='mock-app',
                service_name='mock-service'
            )

    def test_success_default_version(
        self,
        mock_write_working_file,
        mock_run_maven_step,
        mock_determine_container_image_build_tag_info
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app'
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # setup sideeffects
            artifact_parent_dir = os.path.join(test_dir.path, 'target')
            package_artifact_names = [
                f'my-app.jar'
            ]
            def run_maven_side_effect(mvn_output_file_path, step_implementer_additional_arguments):
                os.makedirs(artifact_parent_dir, exist_ok=True)
                for artifact_name in package_artifact_names:
                    artifact_path = os.path.join(
                        artifact_parent_dir,
                        artifact_name
                    )
                    Path(artifact_path).touch()
            mock_run_maven_step.side_effect = run_maven_side_effect

            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:mock-default-version',
                'mock-org/mock-app-mock-service:mock-default-version',
                'localhost',
                'mock-app-mock-service',
                'mock-default-version'
            ]

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='MavenJKubeK8sBuild',
                sub_step_implementer_name='MavenJKubeK8sBuild'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "create container image.",
                name='maven-jkube-output',
                value='/mock/mvn_k8s_build_output.txt'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='mock-default-version',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:mock-default-version',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:mock-default-version',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_k8s_build_output.txt',
                step_implementer_additional_arguments=[
                    '-Djkube.generator.name=localhost/mock-org/mock-app-mock-service:mock-default-version'
                ]
            )
            mock_determine_container_image_build_tag_info.assert_called_once_with(
                image_version=None,
                organization='mock-org',
                application_name='mock-app',
                service_name='mock-service'
            )

    def test_fail_maven_run(
        self,
        mock_write_working_file,
        mock_run_maven_step,
        mock_determine_container_image_build_tag_info
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            step_config = {
                'pom-file': pom_file,
                'organization': 'mock-org',
                'service-name': 'mock-service',
                'application-name': 'mock-app',
                'container-image-version': '1.0-123abc',
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # set up mocks
            mock_determine_container_image_build_tag_info.return_value=[
                'localhost/mock-org/mock-app-mock-service:1.0-123abc',
                'mock-org/mock-app-mock-service:1.0-123abc',
                'localhost',
                'mock-app-mock-service',
                '1.0-123abc'
            ]

            # run step with mock failure
            mock_run_maven_step.side_effect = StepRunnerException('Mock error running maven')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='create-container-image',
                sub_step_name='MavenJKubeK8sBuild',
                sub_step_implementer_name='MavenJKubeK8sBuild'
            )
            expected_step_result.success = False
            expected_step_result.message = \
                "Error running 'maven k8s:build' to create container image. " \
                "More details maybe found in 'maven-jkube-output' report artifact: " \
                "Mock error running maven"
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "create container image.",
                name='maven-jkube-output',
                value='/mock/mvn_k8s_build_output.txt'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='localhost',
                description='Registry URI poriton of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='mock-org',
                description='Organization portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='mock-app-mock-service',
                description='Repository portion of the container image tag' \
                    ' of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='mock-app-mock-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value='1.0-123abc',
                description='Version portion of the container image tag of the built container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='localhost/mock-org/mock-app-mock-service:1.0-123abc',
                description='Full container image tag of the built container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='mock-org/mock-app-mock-service:1.0-123abc',
                description='Short container image tag of the built container image,' \
                    ' excluding the registry URI.'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_once_with(
                mvn_output_file_path='/mock/mvn_k8s_build_output.txt',
                step_implementer_additional_arguments=[
                    '-Djkube.generator.name=localhost/mock-org/mock-app-mock-service:1.0-123abc'
                ]
            )
            mock_determine_container_image_build_tag_info.assert_called_once_with(
                image_version='1.0-123abc',
                organization='mock-org',
                application_name='mock-app',
                service_name='mock-service'
            )



"""`StepImplementer` for the `generate_evidence` step which builds and pushes and artifacts archive.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key           | Required? | Default | Description
----------------------------|-----------|---------|-------------
`organization`              | True      |         | Organization that this workflow is for. \
                                                    Used in archive name.
`application-name`          | True      |         | Application that this workflow is for. \
                                                    Used in archive name.
`service-name`              | True      |         | Service that this workflow is for. \
                                                    Used in archive name.
`version`                   | True      |         | Version of applicaiton service that this \
                                                    workflow is for. \
                                                    Used in archive name. \
`evidence-destination-url` \
                            | False     |         | URL to upload the results archive to.\
                                                    NOTE: TODO describe how URI is formed
                                                    <br /><br />\
                                                    Must start with `/` or `file://` to upload \
                                                    to local file path. \
                                                    <br /><br />\
                                                    Or must start with `http://` or `https://` to \
                                                    upload via a `PUT` to a remote location.
`evidence-destination-username` \
                            | No        |         | Username to use when doing upload via http(s).
`evidence-destination-password` \
                            | No        |         | Password to use when doing upload via http(s).

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key              | Description
---------------------------------|------------
`evidence-path`                  | Absolute path to the generated evidence JSON file.
`evidence-uri`                   | URI of the uploaded results archive.
`evidence-upload-results`        | Results of uploading the results archive to \
                                   the given destination.

"""

import json
import os

from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.file import upload_file

DEFAULT_CONFIG = {

}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'organization',
    'application-name',
    'service-name',
    'version'
]
class GenerateEvidence(StepImplementer):
    """`StepImplementer` for the `report` step which builds and pushes and artifacts archive.
    """

    API_VERSION = "automated-governance/v1alpha1"
    KIND = "WorkflowEvidence"

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Notes
        -----
        These are the lowest precedence configuration values.

        Returns
        -------
        dict
            Default values to use for step configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        #Gather all evidence from steps/sub steps
        evidence = self.__gather_evidence()

        #If no evidence is found then return a step result reflecting that status.
        if not evidence:
            step_result.add_artifact(
                name='result-generate-evidence',
                value='No evidence to generate.',
                description='Evidence from all previously run steps.'
            )
            step_result.message = "No evidence generated from previously run steps"
            return step_result

        org = self.get_value('organization')
        app = self.get_value('application-name')
        service = self.get_value('service-name')
        version = self.get_value('version')

        result_evidence_name = f"{org}-{app}-{service}-{version}-evidence.json"

        work_dir = self.work_dir_path

        evidence_path = os.path.join(work_dir, result_evidence_name)

        #Format into json and write out to file in working directory
        with open(evidence_path, "w", encoding='utf-8') as evidence_file:
            json.dump(evidence, evidence_file, indent=4)

        evidence_destination_url = self.get_value('evidence-destination-url')
        evidence_destination_uri = f"{evidence_destination_url}/" \
                f"{org}/{app}/{service}/{result_evidence_name}"

        #Upload file to datastore
        if evidence_file and evidence_destination_url:
            try:
                upload_result = upload_file(
                    file_path=evidence_path,
                    destination_uri=evidence_destination_uri,
                    username=self.get_value('evidence-destination-username'),
                    password=self.get_value('evidence-destination-password')
                )
                step_result.add_artifact(
                    name='evidence-upload-results',
                    description='Results of uploading the evidence JSON file ' \
                        'to the given destination.',
                    value=upload_result
                )
                step_result.add_artifact(
                    name='evidence-uri',
                    description='URI of the uploaded results archive.',
                    value=evidence_destination_uri
                )

                step_result.success = True
                step_result.message = 'Evidence successfully packaged ' \
                'in JSON file and uploaded to data store.'

            except RuntimeError as error:
                step_result.success = False
                step_result.message = str(error)
                return step_result
        else:
            step_result.success = True
            step_result.message = 'Evidence successfully packaged ' \
            'in JSON file but was not uploaded to data store (no '\
            'destination URI specified).'


        step_result.add_artifact(
            name='evidence-path',
            value=evidence_path,
            description='File path of evidence.'
        )

        return step_result

    def __gather_evidence(self):
        """Gather all collected evidence

        Returns
        -------
        dict
            Dictionary containing evidence (values) for each step (key) that
            has collected evidence.

        """

        gathered_evidence = None
        attestations_keyword = 'attestations'
        workflow_dict = {}

        #Iterate over all previous step results
        for previous_step_result in self.workflow_result.workflow_list:
            step_name = previous_step_result.step_name
            if step_name not in workflow_dict:
                workflow_dict[step_name] = {}

            workflow_dict[step_name][attestations_keyword] = {}

            evidence = previous_step_result.evidence
            for piece_of_evidence in evidence.values():
                workflow_dict[step_name]\
                [attestations_keyword][piece_of_evidence.name]\
                 = {
                    'name': piece_of_evidence.name,
                    'value': piece_of_evidence.value,
                    'description': piece_of_evidence.description,
                 }
                if previous_step_result.environment:
                    workflow_dict[step_name]\
                    [attestations_keyword][piece_of_evidence.name]\
                     ['environment'] = previous_step_result.environment

        #Add step to workflow dictionary
        if workflow_dict:
            gathered_evidence = {
                'apiVersion': GenerateEvidence.API_VERSION,
                'kind': GenerateEvidence.KIND,
                'workflow': workflow_dict
            }

        return gathered_evidence



"""`StepImplementers` for the `create-container-image` step.
"""

from ploigos_step_runner.step_implementers.create_container_image.buildah import \
    Buildah
from ploigos_step_runner.step_implementers.create_container_image.maven_jkube_k8sbuild import \
    MavenJKubeK8sBuild



"""Representation of a configuration value.
"""

import copy
from ploigos_step_runner.decryption_utils import DecryptionUtils

class ConfigValue:
    """Representation of a configuration value.

    Parameters
    ----------
    value : any
        The value of the config option this is the value for.
    parent_source : str file path or dict
        Path to the YML or JSON file that this value is found in or
        the dict that this value is found in.
    path_parts : list
        List of path to the element that this is the value for.

    Attributes
    ----------
    __value : any
        The value of the config option this is the value for.
    __parent_source : str file path or dict
        Path to the YML or JSON file that this value is found in or
        the dict that this value is found in.
    __path_parts : list
        List of path to the element that this is the value for.
    """

    def __init__(self, value, parent_source=None, path_parts=None):
        self.__value = value
        self.__parent_source = parent_source
        self.__path_parts = path_parts

        if self.__path_parts is None:
            self.__path_parts = []

    @property
    def value(self):
        """Get the value of this configuration value (decrypted if applicable).

        If the value happens to be decrypted then will decrypt and then return the decrypted value.

        Returns
        -------
        obj
            Value (decrypted if applicable) of this configuration value.

        See Also
        --------
        raw_value
        """
        # attemp tto decrypt the value
        decrypted_value = DecryptionUtils.decrypt(self)

        # If this value was able to be decrypted, return the decrypted result
        # else return the raw value
        if decrypted_value is not None:
            value = decrypted_value
        else:
            value = self.raw_value

        return value

    @property
    def raw_value(self):
        """Get the value of this configuration value as originally given.

        If the value happens to be encrypted this will return that raw encrypted value as it was
        given to this configuration value.

        Returns
        -------
        obj
            Value of this configuration value as originally given.

        See Also
        --------
        value
        """
        return copy.deepcopy(self.__value)

    @property
    def path_parts(self):
        """Gets copy of the list of path to the element that this is the value for.

        Returns
        -------
        list
            Copy of the list of path to the element that this is the value for.
        """
        return copy.deepcopy(self.__path_parts)

    @property
    def parent_source(self):
        """Get a copy of the source that this configuration value came from.

        Returns
        -------
        str file path or dict
            Path to the YML or JSON file that this value is found in or
            the dict that this value is found in.
        """
        return copy.deepcopy(self.__parent_source)

    def __eq__(self, other):
        """Equality for this object.

        Parameters
        ----------
        other : object
            other object to determine if equal to this object

        Returns
        -------
        bool
            True if this and the other object are both ConfigValue
                objects and have the same value.
            False otherwise.
        """

        if isinstance(other, ConfigValue):
            equal = self.value == other.value
        else:
            equal = False

        return equal

    def __repr__(self):
        """Human readable representation of the object.

        Returns
        -------
        str
            Human readable representation of the object.
        """
        return f"ConfigValue(value={self.raw_value}, value_path='{self.path_parts}')"

    @staticmethod
    def convert_leaves_to_config_values(values, parent_source=None, path_parts=None):
        """In place recursively change all of the leaves of the given
        object to a ConfigValue objects.

        Parameters
        ----------
        values : dict, list, tuple, ConfigValue, None, obj
            Change all the leaves of the given object to ConfigValue objects.
        parent_source : str file path or dict
            Path to the YML or JSON file that this value is found in or
            the dict that this value is found in.
        path_parts : list
            List of path to the element that this is the value for.

        Returns
        -------
        dict, list, None, or ConfigValue
            If given values is a a dict then returns the same dict with all of the leaf values
                changed to ConfigValue or None if already None.
            If given values is list then returns the same list with all of the leaf values
                changes to ConfigValue or None if already None.
            If given values is None, return None.
            If given values is already a ConfigValue return the same ConfigValue

        See Also
        --------
        ConfigValue.convert_leaves_to_config_values
        """
        if path_parts is None:
            path_parts = []

        if isinstance(values, dict): # pylint: disable=no-else-return
            for child_key in values:
                values[child_key] = ConfigValue.convert_leaves_to_config_values(
                    values=values[child_key],
                    parent_source=parent_source,
                    path_parts=(path_parts + [child_key])
                )

            return values
        elif isinstance(values, (list, tuple)):
            for child_key, child_value in enumerate(values):
                values[child_key] = ConfigValue.convert_leaves_to_config_values(
                    values=child_value,
                    parent_source=parent_source,
                    path_parts=(path_parts + [child_key])
                )

            return values
        elif isinstance(values, ConfigValue):
            return values
        elif values is None:
            return None
        else:
            return ConfigValue(
                value=values,
                parent_source=parent_source,
                path_parts=path_parts
            )

    @staticmethod
    def convert_leaves_to_values(values):
        """Recursively transforms all leaves of type ConfigValue to ConfigValue.value

        Parameters
        ----------
        values : dict, list, ConfigValue, or obj
            A collection where the leaves contain ConfigValue to transform back to
            ConfigValue.value

        Returns
        -------
        dict, list, or obj
            If given a dictionary returns that dictionary with all leaves transformed from
                ConfigValue to ConfigValue.value.
            If given a list returns that dictionary with all leaves transformed from
                ConfigValue to ConfigValue.value.
            If given a ConfigValue returns ConfigValue.value
            If any other object returns that object

        See Also
        --------
        ConfigValue.convert_leaves_to_config_values
        """
        if isinstance(values, dict): # pylint: disable=no-else-return
            for child_key in values:
                values[child_key] = ConfigValue.convert_leaves_to_values(values[child_key])

            return values
        elif isinstance(values, (list, tuple)):
            for child_key, child_value in enumerate(values):
                values[child_key] = ConfigValue.convert_leaves_to_values(child_value)

            return values
        elif isinstance(values, ConfigValue):
            return values.value
        else:
            return values



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os
import re
import sys
from unittest.mock import patch

import sh
from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementers.static_code_analysis import \
    SonarQube
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


class TestStepImplementerSonarQubePackageBase(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=SonarQube,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

# TESTS FOR configuration checks
    def test_step_implementer_config_defaults(self):
        defaults = SonarQube.step_implementer_config_defaults()
        expected_defaults = {
            'properties': './sonar-project.properties',
            'java-truststore': '/etc/pki/java/cacerts'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = SonarQube._required_config_or_result_keys()
        expected_required_keys = [
            'url',
            'application-name',
            'service-name',
            'version'
        ]
        self.assertEqual(required_keys, expected_required_keys)

    def test__validate_required_config_or_previous_step_result_artifact_keys_valid(self):
        step_config = {
            'url' : 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
            'application-name': 'app-name',
            'service-name': 'service-name',
            'username': 'username',
            'password': 'password',
            'version': 'notused'
        }

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_invalid_missing_password(self):
        step_config = {
            'url' : 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
            'application-name': 'app-name',
            'service-name': 'service-name',
            'username': 'username',
            'version': 'notused'
        }
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                parent_work_dir_path=parent_work_dir_path
            )
            with self.assertRaisesRegex(
                StepRunnerException,
                r"Either 'username' or 'password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    def test__validate_required_config_or_previous_step_result_artifact_keys_invalid_missing_username(self):
        step_config = {
            'url' : 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
            'application-name': 'app-name',
            'service-name': 'service-name',
            'password': 'password',
            'version': 'notused'
        }
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                parent_work_dir_path=parent_work_dir_path
            )
            with self.assertRaisesRegex(
                StepRunnerException,
                r"Either 'username' or 'password 'is not set. Neither or both must be set."
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_pass(self, sonar_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('sonar-project.properties',b'''testing''')
            properties_path = os.path.join(temp_dir.path, 'sonar-project.properties')

            artifact_config = {
                'version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'properties': properties_path,
                'url': 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                'application-name': 'app-name',
                'service-name': 'service-name',
                'username': 'username',
                'password': 'password'

            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='static-code-analysis',
                sub_step_name='SonarQube',
                sub_step_implementer_name='SonarQube'
            )
            expected_step_result.add_artifact(
                name='sonarqube-result-set',
                value=f'{temp_dir.path}/working/static-code-analysis/report-task.txt'
            )
            expected_step_result.add_evidence(
                name='sonarqube-quality-gate-pass',
                value=True
            )

            sonar_mock.assert_called_once_with(
                    '-Dproject.settings=' + properties_path,
                    '-Dsonar.host.url=https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                    '-Dsonar.projectVersion=1.0-123abc',
                    '-Dsonar.projectKey=app-name:service-name',
                    '-Dsonar.login=username',
                    '-Dsonar.password=password',
                    '-Dsonar.working.directory=' + step_implementer.work_dir_path,
                    _env={'SONAR_SCANNER_OPTS': '-Djavax.net.ssl.trustStore=/etc/pki/java/cacerts'},
                    _out=sys.stdout,
                    _err=sys.stderr
            )

            self.assertEqual(result, expected_step_result)

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_pass_no_username_and_password(self, sonar_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('sonar-project.properties',b'''testing''')
            properties_path = os.path.join(temp_dir.path, 'sonar-project.properties')

            artifact_config = {
                'version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'properties': properties_path,
                'url': 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                'application-name': 'app-name',
                'service-name': 'service-name'

            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='static-code-analysis',
                sub_step_name='SonarQube',
                sub_step_implementer_name='SonarQube'
            )
            expected_step_result.add_artifact(
                name='sonarqube-result-set',
                value=f'{temp_dir.path}/working/static-code-analysis/report-task.txt'
            )
            expected_step_result.add_evidence(
                name='sonarqube-quality-gate-pass',
                value=True
            )

            sonar_mock.assert_called_once_with(
                    '-Dproject.settings=' + properties_path,
                    '-Dsonar.host.url=https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                    '-Dsonar.projectVersion=1.0-123abc',
                    '-Dsonar.projectKey=app-name:service-name',
                    '-Dsonar.working.directory=' + step_implementer.work_dir_path,
                    _env={'SONAR_SCANNER_OPTS': '-Djavax.net.ssl.trustStore=/etc/pki/java/cacerts'},
                    _out=sys.stdout,
                    _err=sys.stderr
            )

            self.assertEqual(result, expected_step_result)

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_fail_no_properties(self, sonar_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            artifact_config = {
                'version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'url': 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                'application-name': 'app-name',
                'service-name': 'service-name'

            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='static-code-analysis',
                sub_step_name='SonarQube',
                sub_step_implementer_name='SonarQube'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Properties file not found: ./sonar-project.properties'

            self.assertEqual(result, expected_step_result)

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_pass_alternate_java_truststore(self, sonar_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('sonar-project.properties', b'''testing''')
            properties_path = os.path.join(temp_dir.path, 'sonar-project.properties')
            temp_dir.write('alternate.jks', b'''testing''')
            java_truststore = os.path.join(temp_dir.path, 'alternate.jks')

            artifact_config = {
                'version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'properties': properties_path,
                'url': 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                'application-name': 'app-name',
                'service-name': 'service-name',
                'username': 'username',
                'password': 'password',
                'java-truststore': java_truststore
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='static-code-analysis',
                sub_step_name='SonarQube',
                sub_step_implementer_name='SonarQube'
            )
            expected_step_result.add_artifact(
                name='sonarqube-result-set',
                value=f'{temp_dir.path}/working/static-code-analysis/report-task.txt'
            )
            expected_step_result.add_evidence(
                name='sonarqube-quality-gate-pass',
                value=True
            )

            sonar_mock.assert_called_once_with(
                    '-Dproject.settings=' + properties_path,
                    '-Dsonar.host.url=https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                    '-Dsonar.projectVersion=1.0-123abc',
                    '-Dsonar.projectKey=app-name:service-name',
                    '-Dsonar.login=username',
                    '-Dsonar.password=password',
                    '-Dsonar.working.directory=' + step_implementer.work_dir_path,
                    _env={'SONAR_SCANNER_OPTS': f'-Djavax.net.ssl.trustStore={java_truststore}'},
                    _out=sys.stdout,
                    _err=sys.stderr
            )

            self.assertEqual(result, expected_step_result)

    def __run__run_step_fail_sonar_scanner_error_test(
        self,
        sonar_scanner_error,
        expected_result_message_regex,
        sonar_mock
    ):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            temp_dir.write('sonar-project.properties',b'''testing''')
            properties_path = os.path.join(temp_dir.path, 'sonar-project.properties')

            artifact_config = {
                'version': {'description': '', 'value': '1.0-123abc'},
            }
            workflow_result = self.setup_previous_result(parent_work_dir_path, artifact_config)

            step_config = {
                'properties': properties_path,
                'url': 'https://sonarqube-sonarqube.apps.ploigos_step_runner.rht-set.com',
                'application-name': 'app-name',
                'service-name': 'service-name',
                'username': 'username',
                'password': 'password'

            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='static-code-analysis',
                implementer='SonarQube',
                workflow_result=workflow_result,
                parent_work_dir_path=parent_work_dir_path,
            )

            sonar_mock.side_effect = sonar_scanner_error

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='static-code-analysis',
                sub_step_name='SonarQube',
                sub_step_implementer_name='SonarQube'
            )
            expected_step_result.success = False
            expected_step_result.add_artifact(
                name='sonarqube-result-set',
                value=f'{temp_dir.path}/working/static-code-analysis/report-task.txt'
            )
            expected_step_result.add_evidence(
                name='sonarqube-quality-gate-pass',
                value=False
            )

            self.assertEqual(result.success, expected_step_result.success)
            self.assertEqual(result.artifacts, expected_step_result.artifacts)
            self.assertRegex(result.message, expected_result_message_regex)

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_fail_sonar_scanner_internal_error(self, sonar_mock):
        self.__run__run_step_fail_sonar_scanner_error_test(
            sonar_scanner_error=sh.ErrorReturnCode_1(
                'sonar-scanner',
                b'mock out',
                b'mock internal error'
            ),
            expected_result_message_regex=re.compile(
                r"Error running static code analysis using sonar-scanner:"
                r".*RAN: sonar-scanner"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock internal error",
                re.DOTALL
            ),
            sonar_mock=sonar_mock
        )

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_fail_sonar_scanner_tests_failed(self, sonar_mock):
        self.__run__run_step_fail_sonar_scanner_error_test(
           sonar_scanner_error=sh.ErrorReturnCode_2(
                'sonar-scanner',
                b'mock out',
                b'mock user error'
            ),
            expected_result_message_regex=re.compile(
                r"Static code analysis failed."
                r" See 'sonarqube-result-set' result artifact for details.",
                re.DOTALL
            ),
            sonar_mock=sonar_mock
        )

    @patch('sh.sonar_scanner', create=True)
    def test_run_step_fail_sonar_scanner_unexpected_error(self, sonar_mock):
        self.__run__run_step_fail_sonar_scanner_error_test(
           sonar_scanner_error=sh.ErrorReturnCode_42(
                'sonar-scanner',
                b'mock out',
                b'mock unexpected error'
            ),
            expected_result_message_regex=re.compile(
                r"Unexpected error running static code analysis using sonar-scanner:"
                r".*RAN: sonar-scanner"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock unexpected error",
                re.DOTALL
            ),
            sonar_mock=sonar_mock
        )



"""DEPRECATED: use ploigos_step_runner.step_implementers.package.MavenPackage instead.

This is an alias to MavenPackage so as to keep things backwards compatible,
but eventually should/can will be removed.
"""

from ploigos_step_runner.step_implementers.package.maven_package import MavenPackage

class Maven(MavenPackage):
    """DEPRECATED: use ploigos_step_runner.step_implementers.package.MavenPackage instead.

    This is an alias to MavenPackage so as to keep things backwards compatible,
    but eventually should/can will be removed.
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment
        )

        print("DEPRECATED: use ploigos_step_runner.step_implementers.package.MavenPackage instead.")




from unittest.mock import patch

from ploigos_step_runner import WorkflowResult
from ploigos_step_runner.step_implementers.package import Maven
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.step_implementers.package.MavenPackage.__init__")
class TestStepImplementerMaven_UnitTestOld___init__(BaseStepImplementerTestCase):
    def test_defaults(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        Maven(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None
        )



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import collections
import functools
import json
import os
import re

from unittest.mock import Mock, patch

from git import Repo
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.generate_metadata import Commitizen


class TestStepImplementerCommitizenGenerateMetadata(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Commitizen,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = Commitizen.step_implementer_config_defaults()
        expected_defaults = {
            'cz-json': '.cz.json',
            'repo-root': './'
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Commitizen._required_config_or_result_keys()
        expected_required_keys = ['cz-json', 'repo-root']
        self.assertEqual(required_keys, expected_required_keys)

    def test__validate_required_config_or_previous_step_result_artifact_keys_valid(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            temp_dir.write('.cz.json', b'''{
                "commitizen": {
                    "bump_message": "build: bump $current_version \\u2192 $new_version [skip-ci]", 
                    "name": "cz_conventional_commits", 
                    "tag_format": "$version", 
                    "update_changelog_on_bump": true
                }
            }''')
            cz_json_file_path = os.path.join(temp_dir.path, '.cz.json')

            step_config = {
                'cz-json': cz_json_file_path,
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Commitizen',
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

    @patch('ploigos_step_runner.step_implementers.generate_metadata.commitizen.sh')
    @patch('ploigos_step_runner.step_implementers.generate_metadata.commitizen.Repo')
    def test_run_step_pass_with_no_existing_tags(self, MockRepo, mock_sh):
        previous_dir = os.getcwd()

        try:
            with TempDirectory() as temp_dir:
                parent_work_dir_path = os.path.join(temp_dir.path, 'working')

                temp_dir.write('.cz.json', b'''{
                    "commitizen": {
                        "bump_message": "build: bump $current_version \\u2192 $new_version [skip-ci]", 
                        "name": "cz_conventional_commits", 
                        "tag_format": "$version", 
                        "update_changelog_on_bump": true
                    }
                }''')
                cz_json_path = os.path.join(temp_dir.path, '.cz.json')

                step_config = {
                    'cz-json': cz_json_path,
                    'repo-root': temp_dir.path
                }
                step_implementer = self.create_step_implementer(
                    step_config=step_config,
                    step_name='generate-metadata',
                    implementer='Commitizen',
                    parent_work_dir_path=parent_work_dir_path,
                )

                mock_sh.cz.bump.side_effect = functools.partial(
                    self._mock_cz_bump, 
                    path=os.path.join(temp_dir.path, '.cz.json'),
                    increment_type='minor'
                )
                MockRepo().tags = []
                result = step_implementer._run_step()

                expected_step_result = StepResult(
                    step_name='generate-metadata',
                    sub_step_name='Commitizen',
                    sub_step_implementer_name='Commitizen'
                )
                expected_step_result.success = True
                expected_step_result.add_artifact(name='app-version', value='0.1.0')

                self.assertEqual(result, expected_step_result)
        finally:
            os.chdir(previous_dir)

    @patch('ploigos_step_runner.step_implementers.generate_metadata.commitizen.sh')
    @patch('ploigos_step_runner.step_implementers.generate_metadata.commitizen.Repo')
    def test_run_step_pass_with_existing_tags(self, MockRepo, mock_sh):
        previous_dir = os.getcwd()

        try:
            with TempDirectory() as temp_dir:
                parent_work_dir_path = os.path.join(temp_dir.path, 'working')
                repo = Repo.init(str(temp_dir.path))

                temp_dir.write('.cz.json', b'''{
                    "commitizen": {
                        "bump_message": "build: bump $current_version \\u2192 $new_version [skip-ci]", 
                        "name": "cz_conventional_commits", 
                        "tag_format": "$version", 
                        "update_changelog_on_bump": true
                    }
                }''')
                cz_json_path = os.path.join(temp_dir.path, '.cz.json')

                step_config = {
                    'cz-json': cz_json_path,
                    'repo-root': temp_dir.path
                }
                step_implementer = self.create_step_implementer(
                    step_config=step_config,
                    step_name='generate-metadata',
                    implementer='Commitizen',
                    parent_work_dir_path=parent_work_dir_path,
                )

                mock_sh.cz.bump.side_effect = functools.partial(
                    self._mock_cz_bump, 
                    path=os.path.join(temp_dir.path, '.cz.json'),
                    increment_type='minor'
                )
                tag = collections.namedtuple('Tag', 'name')
                MockRepo().tags = [
                    tag(name='v_0.1.0__'),
                    tag(name='0.3.0'),
                    tag(name='version_4.34.0'),
                    tag(name='5.2.7-prerelease'),
                    tag(name='v3.142.0-test_abcxyz')
                ]
                result = step_implementer._run_step()

                expected_step_result = StepResult(
                    step_name='generate-metadata',
                    sub_step_name='Commitizen',
                    sub_step_implementer_name='Commitizen'
                )
                expected_step_result.success = True
                expected_step_result.add_artifact(name='app-version', value='5.3.0')

                self.assertEqual(result, expected_step_result)
        finally:
            os.chdir(previous_dir)

    @staticmethod
    def _mock_cz_bump(*args, path, increment_type, **kwargs):
        with open(path, 'r') as cz_json:
            version = json.loads(cz_json.read())['commitizen']['version']
            version = re.match(r'(?P<major>\d+).(?P<minor>\d+).(?P<patch>\d+)', version).groupdict()
            
            if increment_type == 'major':
                version['major'] = str(int(version['major']) + 1)
                version['minor'] = '0'
                version['patch'] = '0'

            if increment_type == 'minor':
                version['minor'] = str(int(version['minor']) + 1)
                version['patch'] = '0'

            if increment_type == 'patch':
                version['patch'] = str(int(version['patch']) + 1)

            kwargs['_out'].write(
                f'tag to create: {version["major"]}.{version["minor"]}.{version["patch"]}'
            )


"""`StepImplementer` for the `package` step using npm

"""
from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.shared.npm_generic import NpmGeneric
from ploigos_step_runner.exceptions import StepRunnerException

class NpmPackage(NpmGeneric):
    """`StepImplementer` for the `package` step using npm.
    """

    def __init__(self, workflow_result, parent_work_dir_path, config, environment):
        super().__init__(
            workflow_result,
            parent_work_dir_path,
            config,
            environment=environment,
            npm_args='install'
        )

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return []

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        npm_output_file_path = self.write_working_file('npm_package_output.txt')
        try:
            self._run_npm_step(
                npm_output_file_path=npm_output_file_path
            )

        except StepRunnerException as error:
            step_result.message = "Error running npm install. See 'npm-output'" \
                f" report artifacts for details: {error}"
            step_result.success = False
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from 'npm install'.",
                name='npm-output',
                value=npm_output_file_path
            )

        return step_result
